#ifndef ENVISION_INCLUDE
#define ENVISION_INCLUDE

#include <elib/list.h>
#include <elib/table.h>

typedef struct cell *cell;
typedef struct instrument *instrument;
typedef struct node *node;
typedef struct edge *edge;
typedef struct space *space;
typedef struct chart *chart;
typedef struct shape_point *shape_point;
typedef struct shape *shape;
typedef struct property_handler *property_handler;
typedef struct property *property;
typedef struct background_line *background_line;


struct property {
  char *name;
  char *value;
};


struct shape_point {
  float x,y;
};

#define MAX_POLYGONAL_DEGREE 12

struct shape {
  unsigned int kind;
  unsigned int degree;
  struct shape_point points[MAX_POLYGONAL_DEGREE];
  struct shape_point edge_locations[MAX_POLYGONAL_DEGREE];
  struct shape_point edge_normals[MAX_POLYGONAL_DEGREE];
};

struct space {
  void *hardware_local;
  cell q;
  void (*draw_node)();
  void (*draw_edge)();
  void (*draw_chart)();
  void (*draw_background_line)();
  void (*snap)();
  int (*redraw)();
  table nodes;
  table edges;
  table default_node_properties;
  table default_edge_properties;
  table node_property_handlers;
  table edge_property_handlers;
  float xscale;
  float yscale;
};

/* these can should also be able to be defined dynamically */
typedef enum node_shape {
    oval_shape,
    hex_shape,
    triangle_shape,
    square_shape,
    ring_shape,
    line_shape,
    none_shape
} node_shape;

 
struct edge {
  node from,to;
  float width;
  instrument i;
  space s;
  int color;

  char *from_tag;
  char *to_tag;
  char *name;
  table properties;
  edge from_next,to_next;
  edge *from_back,*to_back;
  unsigned int from_bin,to_bin;
};

struct node {
  char *name;
  float x,y;
  float width,height,ring_width;
  int color;
  instrument i;
  int selected;
  struct node *parent;
  node children;
  edge froms;
  edge tos;
  space s;
  table properties;
  shape sh;
  node *child_last;
  node child_next;
  unsigned int child_index;
  float *child_weights;
  float rotation;
  float fx,fy;
};

struct chart {
  int which;
  int lines;
  int sample_counter;
  float *state;
  int width,height;
};

struct instrument {
  unsigned int depth;
  void *local;
  int (*probe)(void *,float,float);
  void (*draw)();
  void (*move)(space,void *,float,float);
  float x0,y0,x1,y1;
  cell location;
  struct instrument *next;
  struct instrument **prev;
};

struct property_handler {
  char *name;
  void (*f)();
  void *a;
  property_handler next;
}; 

struct cell {
  cell parent;
  cell children[4];
  int x0,y0,x1,y1;
  instrument instruments;
};

struct background_line {
  unsigned int c;
  unsigned int degree;
  float width;
  float *coordinates;
};

node create_node();
edge create_edge();
background_line create_background_line();
space create_space();
cell create_cell();

instrument insert_instrument(cell q,
			     int (*probe)(void *,float,float),
			     void (*draw)(),
			     void (*move)(space,void *,float,float),
			     void *local,
			     float x0,float y0,float x1,float y1,
			     int depth);
instrument which_instrument(cell q,float x0,float y0);
void foreach_instrument_in(cell q,
			   float x0,float y0,float x1,float y1,
			   unsigned int depth,
			   void (*function)(), void *arg);
void move_instrument(instrument i,float,float,float,float);
void move_node (space,void *,float,float);
extern void coalesce();
extern int position_thunk();
void edge_property();
void node_property();
shape resolve_shape_name();
#ifndef MAX
#define MAX(x,y) (((x)>(y))?(x):(y))
#endif

#ifndef MIN
#define MIN(x,y) (((x)<(y))?(x):(y))
#endif

#endif
