#include <anemone.h>
#include <elib/machine.h>
#include <math.h>

extern char *duplicate_string();

static _inline_ unsigned int string_equal(char *a,char *b)
{
  char *aa=a, *bb=b;

  if (a == b)  return(1);
  if (a && b) {
    while ( (*aa == *bb) && *aa && *bb ) { 
      aa++; 
      bb++;
    } 
    if (*aa==0&&*bb==0) return(1);
  }
  return(0);
}

static int property_compare(property p,char *name)
{
  return(string_equal(p->name,name));
}

/* grouping also */
void edge_position(node local,node remote,float *x,float *y)
{
  if (local->parent) {
    edge_position(local->parent, remote,x,y);
    return;
  } 
  if (remote->parent) {
    edge_position(remote->parent,remote,x,y);
    return;
  }
  switch(local->sh->kind){
  case line_shape:
    {
      float k;
      float w;
      /* this is broken for drawing non-centered remotes */
      if (local->width>local->height){
	k=remote->x-local->x;
	w=local->width/2.0;
	if (k<w) {
	  if (k>-w) *x=remote->x;
	  else *x=local->x-w;
	} else *x=local->x+w;
	*y=local->y;
      } else {
	k=remote->y-local->y;
	w=local->height/2.0;
	if (k<w) {
	  if (k>-w)  *y=remote->y;
	  else *y=local->y-w;
	} else *y=local->y+w;
	*x=local->x;
      }
    }
    break;
  case oval_shape:		
  case ring_shape:
    {
      float nx,ny,d,r;
      nx=remote->x-local->x;
      if (ny=remote->y-local->y){
	*y=sqrt(1.0/((nx*nx/((local->width*local->width/4.0)*ny*ny))+
		     (1.0/(local->height*local->height/4.0))));
	if (ny<0.0) *y*=-1.0;
	*x=(*y)*nx/ny+local->x;
	*y+=local->y;
      } else {
	*y=local->y;
	*x=local->x+(local->width*((nx>0.0)?0.5:-0.5));
      }
    }
    break;
  default:
    *x=local->x;
    *y=local->y;
    break;
  }
}

void set_edge_default(space s,char *prop, char *v)
{
    property p;
    if (!(p=(property)table_find(s->default_edge_properties,prop))){
	p=(property)allocate_memory(sizeof(struct property));
	p->name=duplicate_string(prop);
    } else deallocate_memory(p->value);
    p->value=duplicate_string(v);
    table_insert(s->default_edge_properties,p,prop);
}


void add_edge_property_handler(space s,char *name, void (*f)(), void *a)
{
}


int probe_edge(void *k,float x,float y)
{
  edge e=(edge)k;
  float x0,y0,xn,yn,dn,dt,len;
  
  edge_position(e->from,e->to,&x0,&y0);
  edge_position(e->to,e->from,&xn,&yn);
  xn=xn-x0;
  yn=yn-x0;
  len=sqrt(xn*xn+yn*yn);
  if (len){
    xn/=len;
    yn/=len;
  } else xn=yn=0.0;
  dt=(x-x0)*xn+(y-y0)*yn;
  dn=(x-x0)*yn-(y-y0)*xn;
  if ((fabs(dn) < e->width) && (dt > 0.0) && (dt < len)) return (1);
  return(0);
}


void update_edge_position(edge e)
{
  float x0,x1,y0,y1;
/*
  this is for the adaptive binning
  float x,y,d,d2;	

  d=e->from->x*e->to->sh->edge_normals[e->to_bin].x+
    e->from->y*e->to->sh->edge_normals[e->to_bin].y;		
  d2=e->from->x*e->to->sh->edge_normals[e->to_bin].y+
    e->from->y*e->to->sh->edge_normals[e->to_bin].x;		
  d=e->to->x*e->from->sh->edge_normals[e->from_bin].x+
    e->to->y*e->from->sh->edge_normals[e->from->from-bin].y;		
  d2=e->from->x*e->to->sh->edge_normals[e->to_bin].y+
     e->from->y*e->to->sh->edge_normals[e->to_bin].x		

*/
  /* 
     used to factor in an extra half-node...why??
     if (e->i) move_instrument(e->i,(MIN(e->from->x,e->to->x)-e->width),
     (MIN(e->from->y,e->to->y)-e->width),
     (MAX(e->from->x,e->to->x)+e->width),
     (MAX(e->from->y,e->to->y)+e->width)); 
			    */
  edge_position(e->to,e->from,&x0,&y0);
  edge_position(e->from,e->to,&x1,&y1);

  if (e->i) move_instrument(e->i,
			    MIN(x0,x1),MIN(y0,y1),
			    MAX(x0,x1),MAX(y0,y1));
}

void display_edge(space s,edge e)
{
  float x0,x1,y0,y1;  
  edge_position(e->from,e->to,&x0,&y0);
  edge_position(e->to,e->from,&x1,&y1);

  /*      used to factor in an extra half-node...why??
	  e->i=insert_instrument(s->q,probe_edge,s->draw_edge,0,e,
	  (MIN(e->from->x,e->to->x)-e->width),
	  (MIN(e->from->y,e->to->y)-e->width),
	  (MAX(e->from->x,e->to->x)+e->width),
	  (MAX(e->from->y,e->to->y)+e->width),2);*/
  
  e->i=insert_instrument(s->q,probe_edge,s->draw_edge,0,e,
			 MIN(x0,x1),MIN(y0,y1),
			 MAX(x0,x1),MAX(y0,y1),2);
}

edge create_edge(space s,char *from,char *to,char *label)
{
  edge e;
  property p;
  float *n;

  if (e=table_find(s->edges,label)) return(e);
  e=(edge)allocate_memory(sizeof(struct edge));
  e->from=create_node(s,from);
  e->to=create_node(s,to);
  p=(property)table_find(s->default_edge_properties,"width");
  n=command_parse_float(p->value);
  e->width=*n;
  p=(property)table_find(s->default_edge_properties,"color");
  e->color=xlate_colorspec(p->value);
  e->to_tag=e->from_tag=0;
  display_node(e->from);
  display_node(e->to);
  display_edge(s,e);
  e->s=s;

  if (e->from_next=e->from->froms) e->from_next->from_back=&e->from_next;
  *(e->from_back=&e->from->froms)=e;

  if (e->to_next=e->to->tos) e->to_next->to_back=&e->to_next;
  *(e->to_back=&e->to->tos)=e;

  e->name=duplicate_string(label);
  table_insert(s->edges,e,label);
  
/*  (*s->redraw)(s,e->i->x0,e->i->y0,e->i->x1,e->i->y1,3);*/
  e->properties=create_table(property_compare,key_from_string);
  return(e);
}


void edge_property (space s,char *name, char *prop, char *v)
{
  property p;
  edge e=(edge)table_find(s->edges,name);
  if (e){
    if (!(p=(property)table_find(e->properties,prop))){
      p=(property)allocate_memory(sizeof(struct property));
      p->name=duplicate_string(prop);
    } else deallocate_memory(p->value);
    /*    p->value=duplicate_string(v);
	  table_insert(e->properties,p,name);*/
    if (string_equal(prop,"to_label")){
      e->from_tag=duplicate_string(v);
    }
    if (string_equal(prop,"from_label")){
      e->to_tag=duplicate_string(v);
    }
    if (string_equal(prop,"color")){
      e->color=xlate_colorspec(v);
    } else {
      if (string_equal(prop,"width")) e->width=atof(v);
    }
  }
}

