#include <elib/command.h>
#include <elib/machine.h>

void *command_parse_context(){}
void *command_parse_number_of_arguments(){}
char *command_parse_string(char *s) {return s;}
char *command_print_string(char *s) {return s;}

static int function_compare(command f, char *s)
{
  return(string_equal(f->name,s));
}

static unsigned int key_from_char(unsigned char a)
{
  return(a*1131);
}

static int command_compare(command_type_handler a,unsigned char b)
{
  return(a->desc==b);
}


void command_type(command_space s, char *abbreviation,
		   char *full_name,
		   void *(*translation_function)(), 
		   char *(*print_function)(),
		   void (*deallocate_print)(),
		   void *context)
{
  command_type_handler new=
    (command_type_handler)allocate_memory(sizeof(struct command_type_handler));
  new->desc=abbreviation[0];
  s->handlers[new->desc-32]=new;
  new->name=(char *)duplicate_string(full_name);
  new->conversion_function=translation_function;
  new->print_function=print_function;
  new->context=context;
  new->deallocate_memory=deallocate_print;
}

void command_function(command_space s,
		      void *(*function)(),
		      char *name,
		      char *return_type,
		      char *arguments,
		      void *arg)
{
  command f=(command)allocate_memory(sizeof(struct command));
  f->name=(char *)duplicate_string(name);
  f->return_type=(char *)duplicate_string(return_type);
  f->arguments=(char *)duplicate_string(arguments);
  /*  f->number_of_arguments=string_length(arguments);  */
  f->f=function;
  f->arg=arg;
  table_insert(s->commands,f,name);
}

void command_free_result(char *r)
{
}


static command resolve_command(command_space sp,char *s)
{
  command result;
  if (!sp) return(0);
  if (result=(command)table_find(sp->commands,s)) return(result);
  return(resolve_command(sp->parent,s));
}

command_type_handler resolve_type(command_space sp,char *s)
{
  unsigned int ss=*s;
  command_type_handler c;
  if ((!sp) || (!ss)) return(0);
  if (c=(command_type_handler)sp->handlers[ss-32]) return(c);
  return(resolve_type(sp->parent,s));
}

char *command_print_value(command_space s, void *v, char *t)
{
  command_type_handler ty;
  if (ty=resolve_type(s,t)) return((*ty->print_function)(v,ty->context));
  return(0);
}

char *command_dispatch(command_space s,int argc, char **argv)
{
  void *rval;
  void *arg[8];
  int ac=0;
  int j;
  char *t,*result=0;
  command f=0;
  command_type_handler ty;

  if (argc > 0){
    if (f=resolve_command(s,argv[0])){
      t=f->arguments;
      for (j=1;(*t);){
	if (ty=resolve_type(s,t)){
	  if (ty->conversion_function == command_parse_context) {
	    arg[ac]=f->arg;
	  } else {
	    if (ty->conversion_function  == command_parse_number_of_arguments){
	      arg[ac]=(void *)argc;
	    } else { 
	      arg[ac]=(*ty->conversion_function)(argv[j],ty->context);
	      j++;
	    }
	  }
	  ac++;
	  if ((*(t+1) !='*') || (j>=argc)) t++; 
	} else t++;
      }
    }
    if (f) {
      rval=(*f->f)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],arg[7]);
      result=command_print_value(s,rval,f->return_type);
      for (j=1,ac=0,t=f->arguments;(*t);){
	if (ty=resolve_type(s,t)){
	  if ((ty->conversion_function != command_parse_context)  &&
	      (ty->conversion_function != command_parse_number_of_arguments)){
	    if (ty->deallocate_memory)
	      (*ty->deallocate_memory)(arg[ac],ty->context);
	    j++;
	  }
	  ac++;
	if ((*(t+1) !='*') || (j>=argc)) t++; 
	} else t++;
      }
    } else {
      /*    (*s->error)*/
      printf ("couldn't resolve command %s\n",argv[0]);
    }
  }
  return(result);
}

char *command_string_dispatch (command_space s, char *str)
{
  int something=0;
  char *i,*j,old,*result;
  int argc=0;
  int k;
  char *argv[MAX_COMMAND_ARGUMENTS];
  
  for (i=j=str;*i;i++){
    if ((*i==' ') || (*i=='\t') ||  (*i=='\n')) {
      if (something) argv[argc++]=(char *)duplicate_memory(j,i-j);
      j=i+1;
      something=0;
    } else something=1;
  }
  if (something) argv[argc++]=(char *)duplicate_memory(j,i-j);
  result=command_dispatch(s,argc,argv);
  for (k=0;k<argc;k++) deallocate_memory(argv[k]);
  return(result);
}

void command_line_dispatch(command_space sp, int argc, char **argv)
{ 
  int ss,i,j=0,something=0;
  char **argv2=(char **)allocate_memory(argc*sizeof(char *));
  for (i=1;i<argc;i++){
/*    if (*argv[i]=='-'){ */
    ss=*argv[i];
    if (ss=='-'){ 
      if (something){
	command_dispatch(sp,j,argv2);
	j=0;
      }
      argv2[j++]=argv[i]+1;
    } else {
      argv2[j++]=argv[i];
    }
    something=1;
 }
  if (something) command_dispatch(sp,j,argv2);
  deallocate_memory(argv2);
}

command_space create_command_space(command_space parent,void *ef)
{
  int i;
  command_space new=
    (command_space)allocate_memory(sizeof(struct command_space));
  new->parent=parent;
  new->ef=ef;
  for (i=0;i<96;i++) new->handlers[i]=0;
  new->commands=create_table(function_compare,key_from_string);
  return(new);
}

