#include <elib/table.h>
#include <elib/memory.h>
#include <elib/string.h>
#include <stdarg.h>

static void default_print(char *s){printf (s);}
static table debug_table = (table)0;
static void (*print_function)()=default_print;

void reset_debug()
{
  if (debug_table) {
    iterate_table_entries(debug_table,deallocate_memory);
    destroy_table(debug_table);
    debug_table=create_table(string_equal,key_from_string);
  }
}

void register_debug_print_function(void (*f)())
{
  print_function=f;
}

void set_debug(char *type)
{
  if (debug_table)
    table_insert(debug_table,duplicate_string(type),type);
}

void clear_debug(char *type)
{
  char *f = (char *)0;
  printf("clear_debug(type=%s)\n", type);

  if (debug_table) {
    f = table_find(debug_table,type);
    if (f){
      table_remove(debug_table,type);
      deallocate_memory(f);
    }
  }
}

void debug_init()
{
  debug_table=create_table(string_equal,key_from_string);
}

/*
 * debug() should survive the failure of format_string_internal()
 */
int debug(char *type,char * format,...)
{
  char *z;
  int len;
  int answer = 0;

  va_list a;
  va_start(a,format);
  if (debug_table) {
    if (table_find(debug_table,type)){
      if ((z = format_string_internal(format,&len, a))) {
	(*print_function)(z);
	deallocate_memory(z);
	answer = 1;
      }
    }
  }
  return(answer);
}

