#include <elib/machine.h>
#include <stdarg.h>
#include <elib/command.h>
static command_space print_command_space;

static unsigned int hex_digits[]={'0','1','2','3','4','5','6','7','8','9','a',
				    'b','c','d','e','f'};
#define	char_to_digit(c) (int)((c) - '0')
#define is_digit(c)	 ((char_to_digit(c) >= 0) && (char_to_digit(c) <= 9))
#define	digit_to_char(n) ((n) + '0')
#define isxdigit(x) ( (is_digit(x)) || (((x)>= 'a') && ((x)<='f')) )
#define is_num(x,base) ((base==10||base==8)?is_digit(x):isxdigit(x))
#define isspace(x) \
  ( (x)==' ' || (x)=='\010' || (x)=='\012' || (x)=='\015' )

unsigned int command_parse_hex(char *s)
{
  unsigned int l, done=0;
  unsigned int i, result=0;
  for (l=0;!done;l++){
    if ((i=s[l]))
      result=(result<<4)|((((i)>='0') && ((i)<='9'))?((i)-'0'):
			  (((i)>='a') && ((i)<='f'))?
			  ((i-'a')+10):0);
    else
      done=1;
  }
  return(result);
}

char *command_print_hex(unsigned int x)
{
  unsigned int i=x;
  int k;
  char *result=(char *)allocate_memory(9);
  for (k=7;k>=0;i>>=4,k--) PUT_BYTE(result,k,hex_digits[i&15]);
  PUT_BYTE(result,8,0);
  return(result);
}

int command_parse_decimal(char *s)
{
  int l=0, done=0;
  int negative=0;
  unsigned int i;
  int result=0;
  if (s[0]=='-') l=negative=1;
  for (;!done;l++){
    if ((i=s[l]))
      result=(result*10)+((((i)>='0') && ((i)<='9'))?((i)-'0'):0);
    else
      done=1;
  }
  return(negative?-result:result);
}

float *command_parse_float(char *s)
{
  int negative=0,past=0;
  char *n=s;
  float *result=(float *)allocate_memory(sizeof(float));
  float multiplier=1.0;
  
  *result=0.0;
  if (*n=='-') {
    negative=1;
    n++;
  }
  for (;*n;n++){
    if (*n=='.') past=1;
    else 
      *result=(*result*(past?1:10))+
      ((((*n)>='0') && ((*n)<='9'))?((*n)-'0'):0)*multiplier;
    if (past) multiplier*=0.1;
  }
  *result*=(negative?-1.0:1.0);
  return(result);
}

char *command_print_float(float x)
{
  float i;
  unsigned int k=0;
  int len=0;
  char *result;
  for (i=x;i>0.0;i=i/10,len++);
  result=(char *)allocate_memory(len+2);
  for (i=x,k=len-1;i;i=i/10,k--) PUT_BYTE(result,k,hex_digits[(int)i%10]);
  PUT_BYTE(result,len,0);
  return(result);
}

char *command_print_decimal(int x)
{
  unsigned int i;
  unsigned int k=0;
  int len=0;
  char *result;
  for (i=x;i;i=i/10,len++);
  result=(char *)allocate_memory(len+1);
  for (i=x,k=len-1;i;i=i/10,k--) PUT_BYTE(result,k,hex_digits[i%10]);
  PUT_BYTE(result,len,0);
  return(result);
}

/*
 * Can pass NULL pointers in now.
 * NULL pointers have length zero.
 */
int string_length(char *s)
{
  int result = 0;
  if (s) {
    for (result=0;s[result];result++);
  }
  return(result);
}


void *duplicate_memory(char *x,int len)
{
  void *result=(void *)allocate_memory(len+1);
  bcopy(x,result,len);
  PUT_BYTE(((unsigned char *)result),len,0);
  return(result);
}

/*
 * Duplicating a NULL pointer returns a pointer to
 * a zero length string.  You must call deallocate_memory()
 * on it, as you would with any other duplicate_string() result.
 */
char *duplicate_string(char *x)
{ 
  int len=string_length(x);
  char *result=(char *)allocate_memory(len+1);
  if (x) {
    bcopy(x,result,len+1);
  } else {
    PUT_BYTE(result, 0, 0);
  }
  return(result);
}

/*
 * Two NULL pointers are equal.
 * A NULL pointer is not equal to anything non-NULL.
 * Two non-NULL strings are equal if all of their
 * corresponding bytes are equal.
 */
unsigned int string_equal(char *a,char *b)
{ char *aa=a, *bb=b;
  int answer = 0; /* not equal */

  if (a == b) answer = 1;
  else if (a && b) {
    while ( (*aa == *bb) && *aa && *bb ) { 
      aa++; 
      bb++;
    } 

    if (*aa==0&&*bb==0) answer = 1;
  }
  return(answer);
}

#if 0
unsigned int string_equal(char *a,char *b)
{ char *aa=a, *bb=b;
  while ( (*aa == *bb) && *aa && *bb ) { aa++; bb++; } 
  if (*aa==0&&*bb==0) return(1);
  return(0); /* not equal */
}
#endif

unsigned int key_from_string(char *s)
{
  unsigned int result=0;
  unsigned char *n;
  int i;
  if (!s) return(1);
  for (n=s,i=0;*n;n++,i++) result^=(*n*67869)^(*n*i*513)^(*n*57)^*n*i;
  return(result);
}

void string_copy(char *to, char *from)
{ char *d=to, *s=from;
  for(;*s;) { PUT_BYTE(d,0,*s); d++; s++; }
}

char *string_search(char *scan, char c)
{
  char *s;
  for (s=scan; *s; s++) if ( *s==c) return(s);
  return(0);
}


/* string_format -formats into an internal string then returns a copy 
 * this version has no limit.
 */
#define CHUNK_SIZE 56  /* optimized to fit in 64 bytes(memory.c)*/
#define PUT_BYTE_CHECK(o,i,c) \
{PUT_BYTE(o,i,c); next_out++; \
   if (next_out==CHUNK_SIZE){ \
     outbase->next=(struct chunk *)allocate_memory(sizeof(struct chunk));\
     outbase=outbase->next;\
     outbase->next=0;\
     o=outbase->string;\
     chunks++;\
     next_out=0;\
   }}

/* formats into an internal string then returns a copy */
char *format_string_internal(char *fmt, int *length, va_list ap)
{
  struct chunk {
    char string[CHUNK_SIZE];
    struct chunk *next;
  }output, *outbase;

  int chunks=0;
  char *fc;     /* pointer to current format character */
  char *o;  /* pointer to next output character */
  int qf;       /* flag to quit reading % format */ 
  int zeropad;
  int min_len, len;
  char *c;
  char type;
  char rdigits[12]; /* max negative int with sign + termination */
  unsigned char *X, cX;
  unsigned int x;
  int i,j, base, negative, next_out=0;

  o=output.string;
  output.next=0;
  outbase=(struct chunk *)&output;
  chunks++;

  for (fc=fmt;*fc;fc++){
    if (*fc != '%')
      { PUT_BYTE_CHECK(o,next_out,*fc); }
    else{ 
      min_len=0;
      qf=0;
      zeropad=0;
      fc++;
      if (*fc=='0') { fc++; zeropad=1; }
      while (is_digit(*fc) && *fc){
	min_len=10*min_len + char_to_digit(*fc++);
      }

      type=*fc;
      base =0;
      negative=0;
      switch (type) {

      case '%':
	PUT_BYTE_CHECK(o,next_out,'%');
	break;

      case 's':
	c = va_arg(ap, char *);
	if (!c) c= (char *)"(null)";
	len=string_length(c);
	for (i=0;i<min_len-len;i++){
	  if (zeropad) { PUT_BYTE_CHECK(o,next_out,'0'); }
	  else         { PUT_BYTE_CHECK(o,next_out,' '); }
	}
	for (i=0; i<len; i++) PUT_BYTE_CHECK(o,next_out,*c++); 
	break;

      case 'x':
	base=16;
      case 'o':
	if (!base) base=8;
      case 'u':
	if (!base) base=10;
      case 'd':
      case 'i':
	x = va_arg(ap, unsigned int );
	if (!base){
	  base=10;
	  if ((negative=((int)x <0))) x=(unsigned)(-1*(int)x);
	}
	if (x){
	  for ( i=0; i<12, x; i++, x/=base) 
	    PUT_BYTE(rdigits,i,hex_digits[x%base]);
	} else {
	  PUT_BYTE(rdigits,0,hex_digits[0%base]);
	  i=1;
	}
	if (negative){
	  if (zeropad){
	    PUT_BYTE_CHECK(o,next_out,'-'); 
	    min_len -=1; 
	  }
	  else { PUT_BYTE(rdigits,i,'-'); i++;} 
	}
	len=i;
	for (i=0;i<(min_len-len);i++){
	  if (zeropad) { PUT_BYTE_CHECK(o,next_out,'0');}
	  else         { PUT_BYTE_CHECK(o,next_out,' ');}
	}
	for (i=0; i<len; i++){ 
	  PUT_BYTE_CHECK(o,next_out,rdigits[len-1-i]);
	}
	break;
      case 'X': /* vince special byte to hex string */
	if (min_len==0) min_len=8;
	min_len=(min_len/2)*2;  /* odd number not defined */
	X = va_arg(ap, unsigned char *);
	while (min_len) {
	  PUT_BYTE_CHECK(o,next_out,hex_digits[*X/16]);
	  PUT_BYTE_CHECK(o,next_out,hex_digits[*X&15]);
	  min_len-=2;
	  X++;
	}
	break;
      default:
	break;
      }
    }
  }
  *length=(chunks-1)*CHUNK_SIZE+next_out;
  o= (char *)allocate_memory(*length+1); /* with room for NULL */
  for (i=0,outbase=&output; i<chunks; i++,outbase=outbase->next){
    bcopy(outbase->string, &o[i*CHUNK_SIZE],
	  (i<(chunks-1))?CHUNK_SIZE:next_out);
  }
  PUT_BYTE(o,*length,0);
  for (outbase=output.next;output.next;outbase=output.next){
    output.next=outbase->next;
    deallocate_memory(outbase);
  }
  return(o);
}

void vsprintf(char *out, char *format,...)
{ void *temp;
  int len;
  va_list a;
  va_start(a,format);

  temp=format_string_internal(format,&len, a);
  bcopy(temp,out,len+1);
  va_end(a);
  deallocate_memory(temp);
}

char *string_format(char *format, ...)
{ 
  int len;
  char *result;
  va_list a;
  va_start(a,format);
  result=format_string_internal(format,&len, a);
  va_end(a);
  return(result);
}

/* and for no termination byte */
void format_memory(char *out,char *format,...)
{ char *temp, result;
  int len;
  va_list a;
  va_start(a,format);

  temp=format_string_internal(format,&len, a);
  bcopy(temp,out,len);
  deallocate_memory(temp);
  va_end(a);
}

static int is_string_element(char c, char *scan, char *scan_end)
{ 
  /* char c           character to search for
   * char *scan       where to start the search 
   * char *scan_end   where to stop
  */
  if (*scan_end) PUT_BYTE(scan_end,0,0);
  if ( string_search(scan,c) ) return (1);
  return (0);
}

#define add_char_to_scan(c,s) \
{ if (!is_string_element(*c,scan,s)) { PUT_BYTE(s,0,*c); s++;} c++; }

int vsscanf(char *in, char *fmt,...)
{ char scan[128]; /* inclusion/termination list from %[...]  */
  char *scan_end;
  int scan_exclude;
  int n=0;        /* current poistion of in -- returned by %n */
  int count=0;    /* number of % args translated  - returned on exit*/
  char *fc;       /* pointer to current format character */
  int discard;    /* if not 0; discard parded element */
  int width;      /* optional maximum field width specifier */
  int base;
  int type;
  int sc;
  int in_list;
  int *nn;
  char *s, *c;
  unsigned int *x, xx, cX;
  unsigned char *X;
  va_list ap;
  va_start(ap,fmt);

  for (fc=fmt; *fc; fc++){
    if (*fc != '%'){
      if(*fc == ' '){
	while ( isspace(in[n]) && in[n] ) n++;
      } else {
	if (*fc == in[n]) n++;
	else return (count);
      }
    } else {                                /* *fc == '%' */
      fc++;
      if (*fc=='*') { fc++; discard=1;}
      else { discard=0;}
      width=0;
      while (is_digit(*fc)){
	width = width*10 + char_to_digit(*fc++);
      }
      if (!width) width=-1;
      base=0;
      type=*fc;
      switch (type){

      case '%':
	if (in[n++]!='%') return(count);
	break;


      case 'i':
	base=10;
	while ( isspace(in[n]) && in[n] ) n++;
	if (in[n]=='0') {
	  if (in[n+1]=='x'){
	    base=16;
	    n=n+2;
	  } else {
	    base=8;
	  }
	}
      case 'd':
	if (!base) base=10;
      case 'x':
	if (!base) base=16;
	while ( isspace(in[n]) && in[n] ) n++;
	if (discard) x=&xx;
	else x = va_arg(ap, unsigned int *);
	*x = 0; /* sign? */
	while ( is_num(in[n],base) && width-- && in[n] ) {
	  *x = *x*base + (is_digit(in[n])?char_to_digit(in[n]):10+in[n]-'a');
	  n++;
	}
	if(!discard) count++;
	break;
      case 'X': /* NEW: convert hex string to byte values */
	if (width==-1) width=8;
	width = (width/2)*2; /* behavior for odd width is undefined */
	X = va_arg(ap, unsigned char *);
	while (width && in[n]){
	  if (isxdigit(in[n]))
	    cX = (is_digit(in[n])?char_to_digit(in[n]):10+in[n]-'a');
	  else if (in[n]==' ') cX= 0;
	  else return (count);
	  n++; width--;
	  if (isxdigit(in[n]))
	    cX = cX*16+(is_digit(in[n])?char_to_digit(in[n]):10+in[n]-'a');
	  else if (in[n]==' ') cX= cX*16;
	  else return (count);
	  n++; width--;
	  PUT_BYTE(X,0,cX&255); X++;
	} 
	if(!discard) count++;
	break;

      case 's':
	while ( isspace(in[n]) && in[n] ) n++;
      case 'c':
	if ( type == 'c' && width==-1) width=1;
	if (!discard) s = va_arg(ap, char *);
	while ( !isspace(in[n]) && width-- && in[n] ){
	  if (!discard) { PUT_BYTE(s,0,in[n]); s++; }
	  n++;
	}
	if (!discard && type=='s') PUT_BYTE(s,0,'\0');
	if (!discard) count++;
	break;

      case 'n':
	if (!discard) {
	  nn=va_arg(ap, int *);
	  *nn = n;
	}
	break;

      case '[': 
	fc++;
	scan_end=scan;
	if (*fc=='^'){
	  scan_exclude=1;
	  fc++;
	} else scan_exclude=0;

	/* (] and -) are special if first*/
	if (*fc==']' || *fc=='-'){ 
	  PUT_BYTE(scan_end,0,*fc); 
	  fc++; 
	  scan_end++;
	}
	
	while (*fc != ']' && *fc ) {

	  if (*fc != '-') { add_char_to_scan(fc,scan_end);}
	  else {
	    if (fc[1]==']') { add_char_to_scan(fc,scan_end);}
	    else if (scan_end[-1]<fc[1]){ 
	      for (sc=scan_end[-1]+1; sc<=fc[1];sc++){  
		if (!is_string_element((char)sc,scan,scan_end)) 
		  { PUT_BYTE(scan_end,0,(char)sc); scan_end++; }
	      }
	      fc+=2;
	    }
	    else { add_char_to_scan(fc,scan_end); }
	  } 	
	} 
	if (!*fc) return (count);
	PUT_BYTE(scan_end,0,'\0');
	if (!discard) c=va_arg(ap, char *);
	in_list =(int)string_search(scan, in[n]);
	while ((( scan_exclude &&!in_list) || 
	        (!scan_exclude && in_list)) && in[n]){
	  if (!discard) { PUT_BYTE(c,0,in[n]); c++; }
	  n++;
	  in_list =(int)string_search(scan, in[n]);
	}
	/* if ran into NULL before termination character we are done */ 
	if (scan_exclude && !in[n] ) {
	  PUT_BYTE(c,0,'\0');
	  return (count);
	}
	if (!discard){
	  PUT_BYTE(c,0,'\0'); 
	  count++;
	}
	break;
	  
      default:
	return (count);
	break;
      }
    }
  }
  va_end(ap);
  return(count);
}

