#include <elib/timer.h>
#include <elib/memory.h>

static int sleeping=0;
static int (*block_function)();
static unsigned int (*now)();
static void (*cancel_block)();
     
static int number_of_timers;
static int size_of_pqueue;
static timer timers;
static int beginning_of_epoch;

typedef struct timer_async {
  void (*function)();
  void *arg;
  struct timer_async *next;
} *timer_async;

static timer_async async_pool=0;
static timer_async async_requests=0;
static timer_async *async_request_tail=&async_requests;
static unsigned int timer_set=0;

void register_timer_async()
{
  timer_async new= (timer_async)allocate_memory(sizeof(struct timer_async));
  new->next=async_pool;
  async_pool=new;
}

static int heal(int where)
{
  int left=(where<<1)-1;
  int right=(where<<1);
  int min=where-1;
  struct timer temp;
  
  if (left < number_of_timers)
    if (timers[left].when < timers[min].when) min=left;
  if (right < number_of_timers)
    if (timers[right].when < timers[min].when) min=right;
  if (min != (where-1)){
    temp=timers[where-1];
    timers[where-1]=timers[min];
    timers[min]=temp;
    heal(min+1);
    return(1);
  }
  return(0);
}


void reset_timer(timer t)
{
  t->when=(*now)()+t->interval;
  heal(1);
}

static void add_pqueue(int i)
{
  struct timer temp;
  int me=i-1;
  int parent=(i>>1)-1;
  if (i>1)
    if (timers[me].when< timers[parent].when){
      temp=timers[me];
      timers[me]=timers[parent];
      timers[parent]=temp;
      add_pqueue(i>>1);
    }
}

void remove_timer(timer t)
{
  bcopy(timers+number_of_timers-1,timers+t->index,sizeof(struct timer));
  number_of_timers--;
  heal(1);
}

static timer register_timer_internal(unsigned int when, int interval,
				     int (*function)(),void *argument)
{
  int i;

  if (size_of_pqueue<(number_of_timers+1)){
    size_of_pqueue+=10;
    timers=(timer)realloc(timers,size_of_pqueue* sizeof(struct timer));
  }
  timers[number_of_timers].arg=argument;
  timers[number_of_timers].function=function;
  timers[number_of_timers].interval=interval;
  timers[number_of_timers].when=when;
  add_pqueue(number_of_timers+1);
  return(timers+number_of_timers++);
}

timer register_timer (int interval, int (*function)(),void *argument)
{
  return(register_timer_internal((*now)(),interval,function, argument));
}

timer register_timer_wait (int interval, int (*function)(),void *argument)
{
  return(register_timer_internal((*now)()+interval,interval,function,
				 argument));
}

static int result;
extern int wakeup();

void timer_loop(int block, int interval)
{
  int here,i,ms,cont,poll;
  int finish=(number_of_timers || block)?1:0;

  while (finish){
    poll=0;
    here=(*now)();
    if (number_of_timers){
      if (timers[0].when > here){
	(*block_function)(timers[0].when-here,0);
      } else {
	if ((*timers[0].function)(timers[0].arg)){
	  timers[0].when+=timers[0].interval;
	  heal(1);
	} else remove_timer(timers);
      }
    } else (*block_function)(0,1);
  }
}
  
void initialize_timer(int (*block)(),unsigned int (*time_now)(),
		      void (*cancel_block)())
{
  block_function=block;
  number_of_timers=0;
  now=time_now;
  size_of_pqueue=10;
  timers=(timer)allocate_memory(sizeof(struct timer)*size_of_pqueue);
  gettimeofday(&beginning_of_epoch, 0);
}
