#include <stdarg.h>
#include <anemone.h>
#include <elib/machine.h>
#include <elib/table.h>
#include <protocol/protocol.h>

typedef struct parse_frame *parse_frame;

struct parse_frame { 
    parse_frame parent;
    void (*parent_function)(void *,void *,int,char);
    void *parent_context;
    void (*each_function)(parse_frame *,char);
    void *each_context;
    unsigned int index;
    unsigned int fill_pointer;
    unsigned int size;
    unsigned int count;
};

static _inline_ void char_insert(parse_frame p,char n)
{
  char *z;
  if (!(z=(char *)p->each_context))
    z=(char*)(p->each_context=(void *)allocate_memory(p->size=10));
  if (p->fill_pointer == p->size)
    z=(char*)(p->each_context=
	      (void *)reallocate_memory(p->each_context,p->size+=10));
  z[p->fill_pointer++]=n;
}

void gather_argv(parse_frame *s,char n); 
void parse_string();

typedef struct edge_plist { 
  table t;
  space s; 
} *edge_plist;

typedef struct eprop {
  char *name;
  char *value;
  table t;
} *eprop;

static int eprop_compare(eprop p,char *n)
{
  return(string_equal(p->name,n));
}


static void frame_push(parse_frame *s,
		void (*callback)(void *,void *,int,char),
		void *context,
		void (*each)(parse_frame *,char),
		void *each_context)
{
    parse_frame p=
	(parse_frame)malloc(sizeof(struct parse_frame));
    p->each_function=each;
    p->each_context=each_context;
    p->parent_function=callback;
    p->parent_context=context;
    p->fill_pointer=p->size=0;
    p->parent=*s;
    p->count=0;
    *s=p;
}


static void frame_pop(parse_frame *s, char last)
{
    parse_frame p=*s;
    *s=(*s)->parent;
    if (p->parent_function) (*p->parent_function)(p->parent_context,
						  p->each_context,
						  p->fill_pointer,last);
    free(p);
}



static void eprop_name_finish(eprop p,char *name,int fill,char n)
{
  name[fill]=0;
  p->name=name;
}

static void eprop_value_finish(eprop p,char *value,int fill,char n)
{
  value[fill]=0;
  p->value=value;
  table_insert(p->t,p,p->name);
}

static void gather_word(parse_frame *s,char n)
{
  eprop p;
  switch (n){
  case '=':
    p=(eprop)(*s)->parent_context;
    frame_pop(s,n);
    frame_push(s,eprop_value_finish,p,gather_word,0);
    break;
  case '"':
    frame_pop(s,n);
    break;
  default:
    char_insert(*s,n);
  }
} 

static int free_edge(eprop p,void *foo)
{
  free(p->name);
  free(p->value);
  free(p);
}

static int print_binding (eprop p,void *n)
{
  printf ("%s=%s ",p->name,p->value);
}

static int parse_error (table t,char *field)
{
  printf ("no %s in:",field);
  iterate_table_entries(t,print_binding,0);
  printf ("\n");
}

static void add_damage(char *total, char *this )
{
  if (*total) sprintf(total,"%s, %s",total,this);
  else sprintf(total,"%s",this);
}

float *command_parse_float();

static void ndt_create_edge(edge_plist e,eprop last,int fill, char n)
{
  eprop p,q;
  node s,d;
  float *x,*y;
  float r,g,b;
  char edgename[200];
  char ctemp[50];
  char damages[200];
  char *src=0,*dst=0,*srclat=0,*srclon=0,*dstlat=0,*dstlon=0;
  damages[0]=0;
  if (p=table_find(e->t,"nodename")){
    s=create_node(e->s,p->value);
    if ((p=table_find(e->t,"x")) &&
	(q=table_find(e->t,"y"))){
      x=command_parse_float(p->value);
      y=command_parse_float(q->value);
      move_node(e->s,s,*x,*y);
      deallocate_memory(x);
      deallocate_memory(y);
    }
  } else {
    if (p=table_find(e->t,"srcnode")) {
      src=p->value;
    } else {
      if (p=table_find(e->t,"srcplace")) {
	src=p->value;
      } else {
	if (p=table_find(e->t,"srchost")) {
	  src=p->value;
	} else {
	  if (p=table_find(e->t,"srcip")) {
	    src=p->value;
	  } else {
	    if (p=table_find(e->t,"src_asname")) {
	      src=p->value;
	    } else {
	      if (p=table_find(e->t,"src_as")) 
		src=p->value;
	      else add_damage(damages,"from");
	    }
	  }
	}
      }
    }
    if (p=table_find(e->t,"dstnode")) {
      dst=p->value;
    } else {
      if (p=table_find(e->t,"dstplace")) {
	dst=p->value;
      } else {
	if (p=table_find(e->t,"dsthost")) {
	  dst=p->value;
	} else {
	  if (p=table_find(e->t,"dstip")) {
	    dst=p->value;
	  } else{
	    if (p=table_find(e->t,"dst_asname")) {
	      dst=p->value;
	    } else {
	      if (p=table_find(e->t,"dst_as")) 
		dst=p->value;
	      else  add_damage(damages,"from");
	    }
	  }
	}
      }
    } 
    if (p=table_find(e->t,"srclat")) srclat=p->value;
    else add_damage(damages,"srclat");
    if (p=table_find(e->t,"srclon")) srclon=p->value;
    else add_damage(damages,"srclon");
    if (p=table_find(e->t,"dstlat")) dstlat=p->value;
    else add_damage(damages,"dstlat");
    if (p=table_find(e->t,"dstlon")) dstlon=p->value;
    else add_damage(damages,"dstlon");
    if (src && dst){
      if (srclat && srclon){
	y=command_parse_float(srclat);
	x=command_parse_float(srclon);
	move_node(e->s, create_node(e->s,src),*x,*y);
	deallocate_memory(x);
	deallocate_memory(y);
      }
      if (dstlat && dstlon){
	y=command_parse_float(dstlat);
	x=command_parse_float(dstlon);
	move_node(e->s, create_node(e->s,dst),*x,*y);
	deallocate_memory(x);
	deallocate_memory(y);
      }
      sprintf (edgename,"%s_%s",src,dst);
      if (p=table_find(e->t,"color")){
	sscanf(p->value,"%f %f %f",&r,&g,&b);
	sprintf (ctemp,"%d %d %d",
		 (int)(r*255),
		 (int)(g*255),
		 (int)(b*255));
	create_node(e->s,dst);
	node_property(e->s,dst,"color",ctemp);
      }
      create_edge(e->s,src,dst,edgename);
      if (p=table_find(e->t,"color"))
	edge_property(e->s,edgename,"color",ctemp);
      if (p=table_find(e->t,"width"))
	edge_property(e->s,edgename,"width",p->value);
      
    } else parse_error(e->t,damages);
  }
  iterate_table_entries(e->t,free_edge,0);
  destroy_table(e->t);
}

static void gather_properties(parse_frame *s, char n)
{
  eprop p;
  edge_plist e;
  switch (n){
  case ',':
    break;
  case '"':
    e=(edge_plist)(*s)->each_context;
    p=(eprop)allocate_memory(sizeof(struct eprop));
    p->t=e->t;
    frame_push(s,eprop_name_finish,p,gather_word,0);
    break;
  case ')':
    frame_pop(s,n);
  case ' ':
  case '\n':
    return;
  default:
    printf ("ndt parse error\n");
  }
}

void edge_delimit(parse_frame *s, char r)
{
  table t;
  space sp=(*s)->each_context;
  edge_plist e;
  switch(r){
  case '\n':
  case '\t':
  case ' ':
    break;
  default:
    if (r!='(') {
      printf ("ndt syntax\n");
    } else {
      e=(edge_plist)allocate_memory(sizeof(struct edge_plist));
      e->t=create_table(eprop_compare,key_from_string);
      e->s=sp;
      frame_push(s, ndt_create_edge, e, gather_properties,(void *) e);
    }
  }
}

static void handle_message(layer l, pbuffer b, int f)
{
  parse_frame *g=(parse_frame *)(l->state);
  pbuffer i;
  int j;

  for (i=b;i;i=i->next)
    for (j=0;j<i->length;j++)
      (*(*g)->each_function)(g,i->contents[j]);
  if (f & PROTOCOL_STACK_SHUTDOWN) snap((*g)->each_context);
}

layer create_ndt_layer(space s)
{
  parse_frame *pf=(parse_frame *)allocate_memory(sizeof(parse_frame));
  *pf=0;
  /* wrong that ndt should set these global defaults */
  /*  set_node_default(s,"width","2");
  set_node_default(s,"height","2");
  set_edge_default(s,"width","0.5");*/
  frame_push(pf,0,0,edge_delimit,s);
  return(create_layer(pf,handle_message,handle_message,pf));
}

