#include <anemone.h>
#include <math.h>

static int node_count=0;

static int property_compare(property p,char *name)
{
  return(string_equal(p->name,name));
}

/* real namespace for shapes*/
extern char *duplicate_string();


void set_node_default(space s,char *prop, char *v)
{
    property p;
    if (!(p=(property)table_find(s->default_node_properties,prop))){
	p=(property)allocate_memory(sizeof(struct property));
	p->name=duplicate_string(prop);
    } else deallocate_memory(p->value);
    p->value=duplicate_string(v);
    table_insert(s->default_node_properties,p,prop);
}

void add_node_property_handler(space s,char *name, void (*f)(), void *a)
{
}


/* this has to be shape sensitive*/
int probe_node (void *k, float x, float y)
{
  float xd,yd;
  node n=(node)k;

  if ((n->height > 0.0) && (n->width > 0.0)){
    yd=2.0*(n->y-y)/(float)n->height;
    xd=2.0*(n->x-x)/(float)n->width;
    if ((xd*xd+yd*yd) < 1.0) return (1);
  }
  return(0);
}


void move_node(space sp,void *k,float x,float y)
{
  node n=(node)k;
  float hw=n->width/2+1;
  float hh=n->height/2+1;
  /*  int bx0=MIN(n->x,x)-hw;
      int by0=MIN(n->y,y)-hh;
      int bx1=MAX(n->x,x)+hw;
      int by1=MAX(n->y,y)+hh;*/
  edge e;
  
  n->x=x;
  n->y=y;
  if (n->i){
    move_instrument(n->i,x-hw,y-hh,x+hw,y+hh);
    
    for (e=n->froms;e;e=e->from_next){
      update_edge_position(e);
    }
    for (e=n->tos;e;e=e->to_next){
      update_edge_position(e);
    }
  /*    bx0=MIN(bx0,e->i->x0);
	by0=MIN(by0,e->i->y0);
	bx1=MAX(bx1,e->i->x1);
	by1=MAX(by1,e->i->y1);*/
  /*  (*sp->redraw)(sp,bx0,by0,bx1,by1,3);*/
  }
}


void undisplay_node(node n)
{
  edge e;
  if (n->i){
    remove_instrument(n->i);
    n->i=0;
  }

  for(e=n->froms;e;e=e->from_next) {
    if (e->i) remove_instrument(e->i);
    e->i=0;
  }
  for(e=n->tos;e;e=e->to_next) {
    if (e->i) remove_instrument(e->i);
    e->i=0;
  }
}

void display_node(node n)
{
  edge e;
  float x0,y0,x1,y1;
  if (!n->i){
    n->i=insert_instrument(n->s->q,probe_node,n->s->draw_node,move_node,n,
			   (n->x-n->width/2),(n->y-n->height/2),
			   (n->x+n->width/2),(n->y+n->height/2),1);
    for(e=n->froms;e;e=e->from_next)  if (!e->i) display_edge(n->s,e);
    for(e=n->tos;e;e=e->to_next)   if (!e->i) display_edge(n->s,e);
  }
}

extern float *command_parse_float();

node create_node(space s,char *name)
{
  node n;
  char *rname;
  property p;

  /*  if (!(rname=resolve_alias(name))) rname=name;*/
  /* node indirection */
  rname=name;
  if (n=table_find(s->nodes,rname)) return(n);
  n=(node)allocate_memory(sizeof(struct node));
  /* try to tie this into something* */
  n->x=50+(node_count++%10)*75;
  n->y=50+(node_count/10)*75;
  n->child_last=0;
  n->parent=n->children=0;
  n->selected=0;
  p=(property)table_find(s->default_node_properties,"width");
  /*XX*/
  n->width=*command_parse_float(p->value);
  p=(property)table_find(s->default_node_properties,"height");
  /*XX*/
  n->height=*command_parse_float(p->value);
  p=(property)table_find(s->default_node_properties,"ring_width");
  /*XX*/
  n->ring_width=*command_parse_float(p->value);
  p=(property)table_find(s->default_node_properties,"shape");
  n->sh=resolve_shape_name(p->value);
  p=(property)table_find(s->default_node_properties,"color");
  n->color=xlate_colorspec(p->value);

  n->s=s;
  n->name=duplicate_string(rname);
  table_insert(s->nodes,n,rname);
  /* this should become update_region on bounding box*/
  n->froms=0;
  n->tos=0;
  n->i=0;
    /*  (*s->redraw)(s,n->i->x0,n->i->y0,n->i->x1,n->i->y1,3);*/
  n->properties=create_table(property_compare,key_from_string);
  return(n);
}



void node_property (space s,char *name, char *prop, char *v)
{
  property p;
  node n=(node)table_find(s->nodes,name);
  if (n){
      if (!(p=(property)table_find(n->properties,prop))){
	  p=(property)allocate_memory(sizeof(struct property));
	  p->name=duplicate_string(prop);
      } else deallocate_memory(p->value);
      p->value=duplicate_string(v);
      table_insert(n->properties,p,name);
      if (strcmp(prop,"width")==0) n->width=*command_parse_float(v);
      if (strcmp(prop,"height")==0) n->height=*command_parse_float(v);
      if (strcmp(prop,"shape")==0) n->sh=resolve_shape_name(v);
      if (strcmp(prop,"color")==0) n->color=xlate_colorspec(p->value);
  }
}
