#include <math.h>
#include <anemone.h>
 
typedef struct postscript {
  float x_offset,y_offset,xscale,yscale;
  int orientation;
  space s;
} *postscript;

static int xform(postscript p, int x, int y)
{return(x*p->xscale+p->x_offset); }

static int yform(postscript p, int x, int y)
{ return(y*p->yscale+p->y_offset);}

static int xiform(postscript p, int x, int y)
{ return((x-p->x_offset)/p->xscale); }

static int yiform(postscript p, int x, int y)
{ return((y-p->y_offset)/p->yscale); }

static ps_draw_edge(space sp,edge e)
{
  postscript p=(postscript)sp->hardware_local;
  printf ("%f %f %f setrgbcolor newpath %d %d moveto %d %d lineto\n",
	  (float)((e->color>>16)&255)/255.0,
	  (float)((e->color>>8)&255)/255.0,
	  (float)(e->color&255)/255.0,
	  xform(p,e->from->x,e->from->y),yform(p,e->from->x,e->from->y),
	  xform(p,e->to->x,e->to->y),yform(p,e->to->x,e->to->y));
  printf ("%f setlinewidth stroke\n", e->width*p->xscale);
}

static void local_draw(instrument i,space sp)
{
  (*i->draw)(sp,i->local);
}

static ps_refresh(space sp)
{
  postscript p=(postscript)sp->hardware_local;
  printf ("0 0 0 setrgbcolor newpath 0 0 moveto 612 0 lineto 612 792 lineto 0 792 lineto closepath fill\n");
  /*  printf ("0.5 -0.8 scale\n");	/*  */
  if (p->orientation){
    printf ("90.0 rotate\n");
    printf ("0.5 -0.5 scale\n");
    printf ("608 490 translate\n");
  } else {
    printf ("0.5 -0.5 scale\n");
    printf ("608 -490 translate\n");
  }
printf ("/fontheight 30 def\n");
printf ("/Times-Roman findfont /fontheight load scalefont setfont\n");
printf ("/namedict 12 dict def\n");
printf ("/nodename {\n");
printf (" namedict begin\n");
printf (" /name exch def\n");
printf (" /node_y exch def\n");
printf (" /node_x exch def\n");
printf (" /node_width exch def\n");
printf (" /node_height exch def\n");
printf (" /cut 1 def\n");
printf ("\n");
printf (" {\n");
printf ("   /fontheight load /node_height load div dup mul dup\n");
printf ("   1.0 lt {1.0 exch sub sqrt /node_width load mul} {0} ifelse\n");
printf ("     /name load stringwidth pop lt\n");
printf ("  {\n");
printf ("   /name load dup length dup 3 eq {pop pop exit} if \n");
printf ("   /cut load sub 0 exch getinterval dup length 2 add string 2 copy\n");
printf ("   exch 0 exch putinterval\n");
printf ("   exch length exch dup (..) 4 3 roll exch putinterval \n");
printf ("   /cut 3 def\n");
printf ("   /name exch def\n");
printf ("   pop\n");
printf ("  } { \n");
printf ("   /node_x load /name load stringwidth pop 2 div sub\n");
printf ("    /node_y load /fontheight load 0.25 mul add moveto\n");
printf ("   gsave 1 -1.0 scale /name load show grestore\n");
printf ("   exit\n");
printf ("  } ifelse\n");
printf (" } loop\n");
printf (" end\n");
printf ("} def\n");

  foreach_instrument(sp->q,local_draw,sp);
  printf ("showpage\n");
  exit(0);
}

static ps_redraw(space s)
{
  ps_refresh(s);
}

static ps_resize_space()
{
}

/*#define SCALE_XFER(f) (1.0/pow(0.75,(f-1.0)))*/
#define SCALE_XFER(f) (f)

static ps_draw_node(space sp,node n)
{
  postscript p=(postscript)sp->hardware_local;
  printf ("gsave %f %f %f setrgbcolor %d %d translate gsave %f %f scale \n",
	  (float)((n->color>>16)&255)/255.0,
	  (float)((n->color>>8)&255)/255.0,
	  (float)(n->color&255)/255.0,
	  xform(p,n->x,n->y),yform(p,n->x,n->y),
	  (float)n->width*p->xscale*0.5,(float)n->height*p->yscale*0.5);
  printf ("0 0 1 0 360 arc fill stroke grestore\n");
  printf ("0.5 0.5 0.5 setrgbcolor %f %f 0 0 (%s) nodename grestore\n",
	  1.5*(float)n->height*p->yscale,
	  (float)n->width*p->xscale,
	  n->name);
}

static ps_draw_chart()
{
}

space create_postscript_display(float x,float y,float scale,int orientation)
{
  postscript p=(postscript)allocate_memory(sizeof(struct postscript));
  p->x_offset=x;
  p->y_offset=y;
  p->orientation=orientation;
  p->xscale=SCALE_XFER(scale);
  p->yscale=SCALE_XFER(scale);
  p->s=create_space(p,ps_draw_node,ps_draw_edge,ps_draw_chart,
		    ps_refresh,ps_redraw);
  return(p->s);
}

