#include <protocol/protocol.h>
#include <elib/machine.h>
/*#define BUFFER_POOL */
static pbuffer pool=0;
static int buffer_count=0;
static int buffer_size=0;
static int report=0;
void deallocate_memory();
#ifndef reference_buffer
void reference_buffer(pbuffer p)
{
  p->reference_count++;
}
#endif
pbuffer get_buffer_desc()
{
  pbuffer new;
#ifdef BUFFER_POOL
  if (pool){
    new=pool;
    pool=pool->next;
  } else
#endif
  new =(pbuffer)allocate_memory(sizeof(struct pbuffer));
  return(new);
}
#ifdef BUFFER_POOL
#define get_buffer_desc_local(new) \
  if (pool){ new=pool; pool=pool->next; } \
  else  new =(pbuffer)allocate_memory(sizeof(struct pbuffer))
#else
#define get_buffer_desc_local(new) \
  new =(pbuffer)allocate_memory(sizeof(struct pbuffer))
#endif

void recycle_buffer_desc(pbuffer b)
{
#ifdef BUFFER_POOL
  b->next=pool;
  pool=b;
#else
  deallocate_memory(b);
#endif
}

static void default_free(pbuffer b)
{
  if (b->parent) {
    dereference_buffer(b->parent);
  } else {
    buffer_size-=b->length;
    deallocate_memory(b->contents);
  }
  if (b->down) dereference_buffer(b->down);
  if (b->next) dereference_buffer(b->next);
  recycle_buffer_desc(b);
}

typedef struct buffer_async {
  int size;
  pbuffer b;
  void (*function)();
  int immediate;
  void *arg;
  struct buffer_async *next;
} *buffer_async;

static buffer_async buffer_async_chain;
pbuffer allocate_buffer_async();


/*
 * Finally we have the buffer the caller of 
 * allocate_buffer_async() asked for.  Call the
 * function they told us to and tell them of our success.
 */
static int buffer_async_tail2(void *contents,buffer_async b)
{
  void (*f)()=b->function;
  void *arg=b->arg;
  pbuffer d=b->b;
  d->reference_count=1;
  d->length=b->size;
  d->next=0;
  d->parent=0;
  d->down=0;
  d->free=default_free;
  b->b->contents=(unsigned char *)contents;
  if (!b->immediate){
    b->next=buffer_async_chain;
    buffer_async_chain=b;
    (*f)(d,arg);
  }
return 1;
}

/*
 * Put the new piece of buffer_async_chain into 
 * buffer_async_chain.  Allocate the buffer the
 * caller of allocate_buffer_async() wanted.
 */
static int buffer_async_tail1(pbuffer b,buffer_async d)
{
  d->b=(pbuffer)b;
  return(allocate_memory_async(d->size,buffer_async_tail2,d));
}

/* this all has to be friggin semaphored*/
/*
 * Allocate a piece of buffer_async_chain.
 */
pbuffer allocate_buffer_async(int size,void (*function)(),void *arg)
{
  buffer_async b;
  if (!(b=buffer_async_chain)){
    return(0);
  }
  buffer_async_chain=buffer_async_chain->next;
  b->function=function;
  b->arg=arg;
  b->size=size;
  b->immediate=1;
  if (allocate_memory_async(sizeof(struct pbuffer),buffer_async_tail1,b)){
    b->next=buffer_async_chain;
    buffer_async_chain=b;
    return(b->b);  /* Makes integer from pointer without a cast */
  }
  b->immediate=0;
  return(0);
}

void increment_buffer_pool(size,count)
{
  increment_pool_memory(size,count);
  increment_pool_memory(sizeof(struct pbuffer),count);
}

void register_buffer_async()
{
  buffer_async new;
  register_async_memory();
  register_async_memory();
  register_async_memory();
  register_async_memory(); /* talmage */
  register_async_memory(); /* talmage */
  register_async_memory(); /* talmage */
  new=(buffer_async)allocate_memory(sizeof(struct buffer_async));
  new->next=buffer_async_chain;
  buffer_async_chain=new;
}

/*
 * This ought to check the result of allocate_memory()
 */
pbuffer allocate_buffer(int size)
{
  pbuffer new;
  get_buffer_desc_local(new);
  new->reference_count=1;
  new->length=size;
  new->next=0;
  new->parent=0;
  new->down=0;
  new->contents=(char *)allocate_memory(size);
  new->free=default_free;
  buffer_size+=size;
  return(new);

}

pbuffer append_buffer (pbuffer to,pbuffer from)
{
  pbuffer new=to;
  reference_buffer(from);
  /* this is dangerous semantic erosion by featurism*/
  if (!to) return(from);
  while (new->next) new=new->next;
  new->next=from;
  return(to);
}

pbuffer allocate_fragment(pbuffer b,int offset,int length)
{
  pbuffer f;
  pbuffer result=0;
  pbuffer tail;
  pbuffer step;

  for (f=b;offset>=f->length;f=f->next) offset-=f->length;
  do {
    get_buffer_desc_local(step);
    step->next=0;
    step->contents=f->contents+offset;
    step->length=MIN(length,f->length-offset);
    step->parent=f;
    step->down=0;
    step->free=default_free;
    step->reference_count=0;
    reference_buffer(f);
    offset=0;

    /* result=append_buffer(result(result,step))      SJM */
    if (result) tail->next=step; 
    else            result=step;
    reference_buffer(step); 
    tail=step;

    f=f->next;
  } while ((length-=step->length) > 0);
  return(result);
} 

/* this should actually not* copy in an extremely clever fashion*/
void buffer_copy(pbuffer to, int to_offset, pbuffer from,
		 int from_offset, int length)
{
  int copy_length;
  for (;to_offset>to->length;to=to->next) to_offset-=to->length;
  for (;from_offset>from->length;from=from->next) from_offset-=from->length;
  do{
    copy_length=MIN(to->length-to_offset,from->length-from_offset);
    bcopy(from->contents+from_offset,to->contents+to_offset,copy_length);
    if ((to->length - (to_offset+=copy_length)) == 0){
      to_offset=0;
      to=to->next;
    }
    if ((from->length - (from_offset+=copy_length)) == 0){
      from=from->next;
      from_offset=0;
    }
  } while ((length-=copy_length)>0);
}


void buffer_write(pbuffer b, int offset, char *source, int length)
{ 
  int fflen;
  int source_offset=0;
  
  for (;offset>b->length;b=b->next) offset-=b->length;
  do {
    fflen=MIN(length,b->length-offset);
    bcopy(source+source_offset,b->contents+offset,fflen);
    offset=0;
    source_offset+=fflen;
    b=b->next;
  } while ((length-=fflen) > 0);
}


void buffer_read(pbuffer b, int offset, char *dest, int length)
{
  int fflen, dest_offset=0;

  for (;(offset>b->length);b=b->next) {
    offset-=b->length;
    if (!b->next) { printf("buffer_read: bad pbuffer\n"); break; }
  }
  do {
    fflen=MIN(length,b->length-offset);
    bcopy(b->contents+offset,dest+dest_offset,fflen);
    b=b->next;
    offset=0;
    dest_offset+=fflen;
  } while (((length-=fflen) > 0) && b);
}

int buffer_length(pbuffer b)
{
  pbuffer f=b;
  int result=0;
  for (;f;f=f->next) result+=f->length;
  return(result);
}

void buffer_write_byte (pbuffer b,int offset,unsigned int source)
{
  for (;offset>=b->length;b=b->next) offset-=b->length;
  PUT_BYTE(b->contents,offset,source&255);
}

unsigned char buffer_read_byte(pbuffer b,int offset)
{
  for (;offset>=b->length;b=b->next) offset-=b->length;
  return(*((unsigned char *)b->contents+offset));
}


void buffer_write_long(pbuffer b, int offset, unsigned long source)
{
  for (;offset>=b->length;b=b->next) offset-=b->length;
  if (b->length-offset>3){
    PUT_BYTE(b->contents,offset,source>>24);
    PUT_BYTE(b->contents,offset+1,(source>>16)&255);
    PUT_BYTE(b->contents,offset+2,(source>> 8)&255);
    PUT_BYTE(b->contents,offset+3, source&255);
  } else {
    PUT_BYTE(b->contents,offset,source>>24); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,(source>>16)&255); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,(source>>8)&255); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,source&255);
  }
}

void buffer_write_long_inverted(pbuffer b, int offset, unsigned long source)
{
  for (;offset>=b->length;b=b->next) offset-=b->length;
  if (b->length-offset>3){
    PUT_BYTE(b->contents,offset+3,source>>24);
    PUT_BYTE(b->contents,offset+2,(source>>16)&255);
    PUT_BYTE(b->contents,offset+1,(source>> 8)&255);
    PUT_BYTE(b->contents,offset+0, source&255);
  } else {
    PUT_BYTE(b->contents,offset,source&255); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,(source>>8)&255); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,(source>>16)&255); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
/*     PUT_BYTE(b->contents,offset,(source>>24)source); MWM - 9/15/94 */
    PUT_BYTE(b->contents,offset,(source>>24));
  }
}

void buffer_write_short(pbuffer b,int offset, unsigned short source)
{
  unsigned short temp;
  for (;offset>=b->length;b=b->next) offset-=b->length;
  if (b->length-offset>1){
    PUT_BYTE(b->contents,offset, (source>>8)&255);
    PUT_BYTE(b->contents,offset+1,source&255);
  } else {
    PUT_BYTE(b->contents,offset,source>>8); offset++;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,source&255);
  }
}

unsigned long buffer_read_long( pbuffer b, int offset)
{
  unsigned int result=0;
  for (;offset>=b->length;b=b->next) offset-=b->length;
  if (b->length-offset>3){
    return(result=
	   (b->contents[offset  ]<<24)|
	   (b->contents[offset+1]<<16)|
	   (b->contents[offset+2]<< 8)|
	   (b->contents[offset+3]));
  } else {
    result=b->contents[offset++]<<24;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    result|=(b->contents[offset++]<<16);
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    result|=(b->contents[offset++]<<8);
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    result|=b->contents[offset];
  }
  return(result);
}

unsigned short buffer_read_short(pbuffer b,int offset)
{
  unsigned short result=0;
  for (;offset>b->length;b=b->next) offset-=b->length;
  if (b->length-offset>1){
    return(result=b->contents[offset]<<8|b->contents[offset+1]);
  } else {
    result=b->contents[offset++]<<8;
    if (b->length == offset){
      b=b->next;
      offset=0;
    }
    result|=b->contents[offset];
  }
  return(result);
}

void buffer_write_zero(pbuffer b,int offset, int length)
{
  for (;offset>=b->length;b=b->next) offset-=b->length;
  for (;length;length--,offset++) {
    if (offset >= b->length) {
      b=b->next;
      offset=0;
    }
    PUT_BYTE(b->contents,offset,0);  
  }
}


#undef dereference_buffer
void dereference_buffer(pbuffer p)
{
  if ((--p->reference_count)==0){
    (*p->free)(p);
  } else {
    if (p->reference_count < 0)
      printf ("buffer management error\n");
  }
}

