#include <protocol/protocol.h>
#include <elib/list.h>

typedef struct framer {
  pbuffer queue;
  int state;
  int total_size;
  int queue_size;
} *framer;

static void framer_down(layer l, pbuffer b, int f)
{
  int len=buffer_length(b)-4;
  buffer_write(b,0,&len,4);
  write_down(l,b,f);
}

static void framer_up(layer l, pbuffer b, int f)
{
  framer fr=(framer)l->state;
  int fragment_length;
  pbuffer old_queue;
  int change=1;

  if (b){
    fragment_length=buffer_length(b);
    fr->queue=append_buffer(fr->queue,
			    old_queue=allocate_fragment(b,0,fragment_length));
    dereference_buffer(old_queue);
    fr->queue_size+=fragment_length;
    
    while (change){
      change=0;
      if ((!fr->state) && (fr->queue_size > 4)){
	buffer_read(fr->queue,0,&fr->total_size,4);
	fr->state=1;
      }
      if (fr->state && (fr->queue_size >= (fr->total_size+4))){
	write_up(l,old_queue=allocate_fragment(fr->queue,4,fr->total_size),0);
	dereference_buffer(old_queue);
	fr->state=0;
	fr->queue_size-=(fr->total_size+4);
	old_queue=fr->queue;
	if (fr->queue_size){
	  fr->queue=allocate_fragment(fr->queue,4+fr->total_size,
				      fr->queue_size);
	}	else {
	  fr->queue=0;
	}
	dereference_buffer(old_queue);
	change=1;
      }
    }
  }
  if (f & PROTOCOL_CONNECTED) write_up(l,0,f);
}

static void framer_fixed_up(layer l, pbuffer b, int f)
{
  framer fr=(framer)l->state;
  int fragment_length;
  pbuffer old_queue;

  if (b){
    fragment_length=buffer_length(b);
    fr->queue=append_buffer(fr->queue,
			    old_queue=allocate_fragment(b,0,fragment_length));
    dereference_buffer(old_queue);
    fr->queue_size+=fragment_length;
    while (fr->queue_size >= fr->total_size){
      write_up(l,old_queue=allocate_fragment(fr->queue,0,fr->total_size),0);
      dereference_buffer(old_queue);
      fr->queue_size-=(fr->total_size);
      old_queue=fr->queue;
      fr->queue= (fr->queue_size)?
	allocate_fragment(fr->queue,fr->total_size,fr->queue_size):0;
      dereference_buffer(old_queue);
    }
  }
  if (f & PROTOCOL_CONNECTED) write_up(l,0,f);
}

static pbuffer framer_allocate_down(l,s)
{
  pbuffer b=allocate_downwards_buffer(l,s+4);
  pbuffer result=allocate_fragment(b,4,s);
  result->down=b;
  return(result);
}

layer create_framer(int size)
{
  framer new=(framer)allocate_memory(sizeof(struct framer));
  new->state=0;
  new->queue=0;
  new->queue_size=0;
  if (size) {
    new->total_size=size;
    return(create_layer(new,0,framer_fixed_up,0));
  } else {
    return(create_layer(new,framer_down,framer_up,framer_allocate_down));
  }
}
