#include <protocol/protocol.h>

typedef struct frob {
  int xmit_count;
  int recv_count;
  int frequency;
  layer l;
  int min_size,max_size;
  int which_size;
  int timer;
  int sequence_errors;
  int payload_errors;
  int report;
} *frob;

static int ack(frob f)
{
  int spread=f->max_size-f->min_size;
  int len=(f->which_size=
	   spread?((f->which_size-f->min_size+1)%spread)+
	   f->min_size:f->max_size);
  pbuffer message=allocate_downwards_buffer(f->l,len);
  int i;
  unsigned int e;
  
  for (i=4;i<len;i++){
    e=((i+1)*f->xmit_count)&255;
    buffer_write_byte(message,i,e);
  }
  buffer_write_long(message,0,f->xmit_count);
  f->xmit_count++;
  if (f->l->down) write_down(f->l,message,0);
  if (f->l->up) write_up(f->l,message,0);
  dereference_buffer(message);
}

static void top(layer l, pbuffer b, int flags)
{
  frob f=(frob)l->state;
  int length;
  int c;
  int i;
  unsigned char k;

  if (b){
    length=buffer_length(b);
    c=buffer_read_long(b,0);
    if (c != f->recv_count){
      printf ("SEQUENCE ERROR got: %d expected: %d\n",c,f->recv_count);
      f->sequence_errors++;
      f->recv_count=c;
    } else {
      for (i=4;i<length;i++){
	buffer_read(b,i,&k,1);
	if (k != (((i+1)*c) % 256)) {
	  printf ("payload error \n");
	  f->payload_errors++;
	  i=length;
	}
      }
    }
    f->recv_count++;
    if (!(f->recv_count%f->report))
      printf ("sent: %d recv: %d sequence: %d payload: %d\n",
	      f->xmit_count,f->recv_count,f->sequence_errors,
	      f->payload_errors);
  }
  if ((flags & PROTOCOL_CONNECTED) && f->frequency){
    printf ("frob connected\n");
    f->timer=register_timer (f->frequency,ack,f);
    if (f->l->up) write_up(f->l,0, PROTOCOL_CONNECTED);
  }
  if (flags & PROTOCOL_STOP){
    printf ("stop\n");
    if (f->timer)remove_timer(f->timer);
    f->timer=0;
  }
  if ((flags & PROTOCOL_START) && f->frequency) {
    printf ("start\n");
    if (!f->timer) 
      f->timer=register_timer (f->frequency,ack,f);
  }
}

layer create_frob(int frequency,int min_size,int max_size, int report)
{
  frob new=(frob)allocate_memory(sizeof(struct frob));
  new->frequency=frequency;
  new->max_size=max_size;
  new->min_size=min_size;
  new->which_size=new->min_size;
  new->xmit_count=new->recv_count=0;
  new->report=report;
  new->sequence_errors=0;
  new->payload_errors=0;
  return(new->l=create_layer(new,top,top,0));
}

