#include <protocol/protocol.h>
#include <elib/machine.h>
#include <elib/table.h>
#define MAX_HEADER 8

typedef struct llc{
  layer l;
  table protocols;
  int connected;
} *llc;

typedef struct llc_protocol{
  layer l;
  llc lc;
  int header_length;
  unsigned char *header;
} *llc_protocol;

static int compare_llc(llc_protocol l,char *id)
{
  return(bcmp(l->header,id,l->header_length)==0);
}


static int llc_key(unsigned char *a)
{
  int i;
  int result=0;
  for (i=0;i<MAX_HEADER;i++) result=(result<<5)^a[i];
  return(result);
}

static pbuffer allocate_downwards(layer l,int s)
{
  llc_protocol p=(llc_protocol)l->state;
  int total=s+p->header_length;
  pbuffer b=allocate_downwards_buffer( p->lc->l,total);
  pbuffer d=allocate_fragment(b,p->header_length,s);
  d->down=b;
  buffer_write(b,0,p->header,p->header_length);
  return(d);
}

static int send_flags(llc_protocol p,unsigned int f)
{
  write_up(p->l,0,f);
}

static void protocol_up(layer l,pbuffer b,int f)
{
  llc_protocol p=(llc_protocol)l->state;
  if ((f & PROTOCOL_CONNECTED) && p->lc->connected)
    write_up(l,b,f);
}

static void down(layer l,pbuffer b,int f)
{
  llc_protocol p=(llc_protocol)l->state;
  write_down(p->lc->l,b,f);
}

static void up(layer l,pbuffer b,int f)
{
  unsigned char header[MAX_HEADER];
  llc_protocol p;
  llc ll=(llc)l->state;
  int i;
  pbuffer d;

  for(i=0;i<MAX_HEADER;i++) PUT_BYTE(header,i,0);
  if (f&PROTOCOL_CONNECTED){
    ll->connected=1;
    iterate_table_entries(ll->protocols,send_flags,PROTOCOL_CONNECTED);
  }
  if (f&PROTOCOL_DISCONNECTED){
    ll->connected=0;
    iterate_table_entries(ll->protocols,send_flags,PROTOCOL_DISCONNECTED);
  }
  if (f&PROTOCOL_STACK_SHUTDOWN)
    iterate_table_entries(ll->protocols,send_flags,PROTOCOL_STACK_SHUTDOWN);

  if (b){
    buffer_read(b,0,header,3);
    if ((header[0]==0xaa) && (header[1]==0xaa)  && (header[2]==0x03)){
      buffer_read(b,3,header+3,5);
      if (p=(llc_protocol)table_find(ll->protocols,header)){
	write_up(p->l,
		 d=allocate_fragment(b,p->header_length,
				     buffer_length(b)-p->header_length),f);
	dereference_buffer(d);
      } else {
	printf ("%s\n",string_format("unbound llc-snap protocol %016X",
				     header));
      }
    }
  }
}


layer add_llc_snap(layer l,unsigned int oui,unsigned int pid)
{
  int i;
  llc_protocol new=(llc_protocol)allocate_memory(sizeof(struct llc_protocol));
  new->lc=(llc)(l->state);
  new->header_length=8;
  new->header=(unsigned char *)allocate_memory(MAX_HEADER);
  for(i=0;i<MAX_HEADER;i++) PUT_BYTE(new->header,i,0);
  PUT_BYTE(new->header,0,0xaa);
  PUT_BYTE(new->header,1,0xaa);
  PUT_BYTE(new->header,2,0x03);
  PUT_BYTE(new->header,3,oui>>16);
  PUT_BYTE(new->header,4,(oui>>8)&255);
  PUT_BYTE(new->header,5,oui&255);
  PUT_BYTE(new->header,6,pid>>8);
  PUT_BYTE(new->header,7,pid&255);
  table_insert(new->lc->protocols,new,new->header);
  return(new->l=create_layer(new,down,protocol_up,allocate_downwards));
}

layer create_llc()
{
  llc new=(llc)allocate_memory(sizeof(struct llc));
  new->protocols=create_table(compare_llc,llc_key);
  new->connected=0;
  return(new->l=create_layer(new,0,up,0));
}


