#include <protocol.h>
#include <list.h>


typedef struct little_mux *little_mux;

typedef struct mux {
  pbuffer **queue;
  int receive;
  int lowest_expected;
  int highest_expected;
  layer from_above;
  little_mux muxes;
  int rank;
  int *maps;
  int mask;
} *mux;

struct little_mux {
  mux m;
  layer l;
  int index;
  int sequence_number;
};

static void mux_down(layer l, pbuffer b, int f)
{
  mux m=(mux)l->state;
  int len=buffer_length(b);
  int offset,fragment_size=len/m->rank;
  pbuffer temp;
  int transfer_size;
  int i;

  for (i=0;i<m->rank;i++){
    buffer_write(b,offset,&i,4);
    transfer_size= MIN(fragment_size,len-i*(fragment_size));
    write_down((m->muxes[i]).l,
	       temp=allocate_fragment(b,offset,transfer_size),f);
    dereference_buffer(temp);
  }
}

static void mux_up(layer l, pbuffer b, int f)
{
  little_mux lm=(little_mux)l->state;
  mux m=lm->m;
  int fragment_length;
  int buffer_index;
  int i;
  pbuffer aggregate=0;
  pbuffer fragment,temp;

  buffer_index=lm->sequence_number-m->lowest_expected;
  if (buffer_index > m->highest_expected) {
    m->maps=(int *)realloc(m->maps+m->receive,
				(m->highest_expected-m->receive)
				*sizeof(int));
    m->queue=(pbuffer **)realloc(m->maps+m->receive,
				     (m->highest_expected-m->receive)*
				     sizeof(pbuffer));
    m->highest_expected++;
    m->lowest_expected=m->receive;
  }
  
  *(*(m->queue+buffer_index)+lm->index)=b;
  lm->sequence_number++;
  reference_buffer(b);
  if ((m->maps[buffer_index] |= (1<<lm->index)) == m->mask) {
    for (i=0;i<m->rank;i++){
      fragment=*(*(m->queue+buffer_index)+i);
      temp=allocate_fragment(fragment,4,buffer_length(fragment)-4);
      aggregate=append_buffer(aggregate,temp);
      dereference_buffer(fragment);
      dereference_buffer(temp);
    }
    write_up(m->from_above,aggregate,0);
    m->receive++;
    dereference_buffer(aggregate);
  }
}

list create_mux(int number)
{
  mux new=(mux)allocate_memory(sizeof(struct mux));
  int i;
  layer from_above=create_layer(new,mux_down,0,0);
  list result;
  
  new->from_above=from_above;
  result=insert(from_above,0);
  new->muxes=(little_mux)allocate_memory(number*sizeof(struct little_mux));
  new->maps=(int *)sizeof(number*sizeof(int));
  new->mask=0;
  new->queue=(pbuffer **)allocate_memory(number*sizeof(pbuffer));
  new->lowest_expected=new->highest_expected=new->receive=0;
  for (i=0;i<number;i++) {
    new->mask|=(1<<i);
    new->muxes[i].m=new;
    new->muxes[i].l=create_layer(new->muxes+i,0,mux_up,0);
    result=insert(new->muxes[i].l,result);
  }
  return(result);
}
