#ifndef ATM_PROTOCOLS
#define ATM_PROTOCOLS

#define PROTOCOL_END_OF_MESSAGE 1
#define PROTOCOL_BEGINNING_OF_MESSAGE 2
#define PROTOCOL_ASSURED_DELIVERY 4
#define PROTOCOL_STACK_SHUTDOWN 8
#define PROTOCOL_CRC_INVALID 16
#define PROTOCOL_HARDWARE_CRC10 32
#define PROTOCOL_CONNECTED 64
#define PROTOCOL_DISCONNECTED 128
#define PROTOCOL_PRIORITY 256
#define PROTOCOL_PRINT 512
#define PROTOCOL_HARDWARE_CRC32 1024
#define PROTOCOL_ATM_RAW 2048
#define PROTOCOL_STOP 4096
#define PROTOCOL_START 8192

/*
#define PROTOCOL_END_OF_MESSAGE (1 | (1<<30))
#define PROTOCOL_BEGINNING_OF_MESSAGE (2 | (1<<30))
#define PROTOCOL_ASSURED_DELIVERY (4 | (1<<30))
#define PROTOCOL_STACK_SHUTDOWN (8 | (1<<30))
#define PROTOCOL_CRC_INVALID (16 | (1<<30))
#define PROTOCOL_HARDWARE_CRC10 (32 | (1<<30))
#define PROTOCOL_CONNECTED (64 | (1<<30))
#define PROTOCOL_DISCONNECTED (128 | (1<<30))
#define PROTOCOL_PRIORITY (256 | (1<<30))
#define PROTOCOL_PRINT (512 | (1<<30))
*/

#define LLC_VC_TIMEOUT  900   /* in seconds, 15 minutes */

typedef struct pbuffer *pbuffer;
typedef struct layer *layer;
typedef struct fragment *fragment;

#define PAD_LENGTH(size,padto) (((size)%(padto))?((padto)-((size)%(padto))):0)

struct layer{
  void (*receive_down)();
  void (*receive_up)();
  pbuffer (*allocate_down)();
  void *state;
  layer up;
  layer down;
};


struct pbuffer {
  pbuffer parent;
  int reference_count;
  unsigned char *contents;
  pbuffer next;
  int length;
  pbuffer down;
  void (*free)();
};

void print_layer();

layer create_layer();
layer create_aal4();
layer create_aal5();
layer create_errors();
layer create_print();
layer create_sscop();
layer create_udp();
layer create_llc();
layer create_special_file();
layer create_frob();
layer create_filedesc();
layer create_chain();
layer add_chain();

pbuffer allocate_downwards_buffer();
pbuffer allocate_buffer();
pbuffer allocate_fragment();
pbuffer append_buffer();
pbuffer allocate_buffer_async();

void reference_buffer();
void dereference_buffer();

#ifndef MAX
#define MAX(x,y) ((x)>(y)?(x):(y))
#endif

#ifndef MIN
#define MIN(x,y) ((x)<(y)?(x):(y))
#endif
pbuffer get_buffer_desc();
unsigned char buffer_read_byte();
unsigned short buffer_read_short();
unsigned long buffer_read_long();

#define dereference_buffer(p)                    \
 if ((--(p)->reference_count)==0) (*(p)->free)((p))
/*  else if ((p)->reference_count < 0) printf ("buffer management error\n");}*/
#define reference_buffer(p)  ((p)->reference_count++)

#define write_up(l,b,flags)                                     \
{ if (l){                                                       \
    if ((l)->up){                                               \
      if ((l)->up->receive_down)                                \
        (*(l)->up->receive_down)((l)->up,(b),(flags));          \
      else                                                      \
        write_up_internal((l)->up,(b),(flags));                 \
    }                                                           \
  }                                                             \
}

#define write_down(l,b,flags)                                  \
{ pbuffer DONT_USE_THIS=b;                                     \
   if (l){                                                     \
    if ((l)->down){                                            \
      if ((l)->down->receive_up)                               \
        (*(l)->down->receive_up)((l)->down,DONT_USE_THIS?DONT_USE_THIS->down:0,(flags));     \
      else                                                     \
        write_down_internal((l)->down,DONT_USE_THIS?DONT_USE_THIS->down:0,(flags));        \
    }                                                          \
  }                                                            \
 }


#endif
