#include <protocol/protocol.h>
#include <elib/string.h>  /* format_string_internal() */
#include <stdarg.h>


void print_layer(layer l)
{
  printf("l= %x\n", l);
  if (l != (layer)0) {
    printf("  l->receive_down= %x l->receive_up=%x l->allocate_down=%x\n",
	   l->receive_down, l->receive_up, l->allocate_down);
    printf("  l->state=%x l->up= %x l->down=%x\n",
	   l->state, l->up, l->down);
  }
}


void write_up_internal(layer l,pbuffer b,int flags)
{
  write_up(l,b,flags);
}

void write_down_internal(layer l,pbuffer b,int flags)
{
  write_down(l,b,flags);
}

void write_string_down(layer l,int flags,char* format,...)
{
  char *temp;
  pbuffer b;
  int len;
  va_list a;
  va_start(a,format);
  temp=(char *)format_string_internal(format,&len,a);
  b=allocate_downwards_buffer(l,len);
  buffer_write(b,0,temp,len);
  write_down(l,b,flags);
  dereference_buffer(b);
  deallocate_memory(temp);
}

layer create_layer(void *state,void (*receive_up)(),
		   void (*receive_down)(), pbuffer (*allocate_down)())
{
  layer new=(layer)allocate_memory(sizeof(struct layer));
  new->state=state;
  new->receive_up=receive_up;
  new->receive_down=receive_down;
  new->allocate_down=allocate_down;
  new->up=new->down=0;
  return(new);
}

pbuffer allocate_downwards_buffer(layer l,int size)
{
  pbuffer r1,r2;;
  if (!l->down) return(allocate_buffer(size));
  if (l->down->allocate_down){
    return((*l->down->allocate_down)(l->down,size));
  } else {
    r1=allocate_downwards_buffer(l->down,size);
    r2=allocate_fragment(r1,0,size);
    r2->down=r1;
    return(r2);
  }
}

static layer assemble_internal (int total,layer last,va_list frame,
                                int count,layer *bottom)
{
  layer this=0;
  if (count < total){
    this=va_arg(frame,layer);
    this->up=last;
    this->down=assemble_internal(total,this,frame,count+1,bottom);
  } else {
    *bottom=last;
  }
  return(this);
}

void staple(layer top,layer bottom)
{
  top->down=bottom;
  bottom->up=top;
}

void assemble_stack (int number,...)
{
  va_list a;
  layer bottom;
  va_start(a,number);
  assemble_internal(number,0,a,0,&bottom);
  if (bottom->receive_down)
    (*bottom->receive_down)(bottom,0,PROTOCOL_CONNECTED);
  else write_up(bottom,0,PROTOCOL_CONNECTED);
  va_end(a);
}

