/*
 * File:  skip_parse.c
 * Author: 
 * Version: $Id: skip_parse.c,v 1.2 1996/07/01 01:52:25 yuri Exp $
 * Purpose: 
 */
#ifndef __LINT__
static char rcsid[] = "$Id: skip_parse.c,v 1.2 1996/07/01 01:52:25 yuri Exp $";
#endif

#include <elib/machine.h>
#include <elib/list.h>
#include <elib/string.h>
#include <elib/memory.h>
#include <stdarg.h>
#include <skip/skip.h>
#include <elib/rmap.h>

/* could use a global symbol table*/
static rmap gensym_counter;
environment skip_global_environment;
lambda create_lambda();
int skip_register_timer();
extern skip_load_file();
char *skip_eval_string();
void lambda_dispatch();

int compare_variable(char **a,char *b)
{
  return(string_equal(a[0],b));
}

static _inline_ char *append_string(parser_frame p,int argc,char **argv)
{
  char *new;
  int lookaside_len[10];
  int i,len=0,n;

  for (i=0;i<argc;i++)  {
    len+=n=string_length(argv[i]);
    if (argc<10) lookaside_len[i]=n;
  }
  if (p->free_length < len){
    p->argv[p->argc]=reallocate_memory(p->argv[p->argc],
				       p->fill_pointer+len+10);
    p->free_length=len+10;
  }
  for (i=0;i<argc;i++){
    if (i<10) len=lookaside_len[i];
    else len=string_length(argv[i]);
    bcopy(argv[i],p->argv[p->argc]+p->fill_pointer,len);
    p->fill_pointer+=len;
    p->free_length-=len;
  }
  return(0);
}

static char *skip_if(environment e,int args,int a,char *b,char *c)
{
  if (a){
    return(skip_eval_string(e,b));
  } else{
    if (args > 3) return(skip_eval_string(e,c));
  }
  return(0);
}

static char *resolve_variable(environment i,int c,char **a)
{
  char **v;
  if (i){
    if (v=(char **)table_find(i->variables,a[0]))  return(v[1]);
    return(resolve_variable(i->parent,c,a));
  }
  printf ("couldn't resolve variable %s\n",a[0]);
  return(0);
}

void set(environment i,char *a,char *b)
{
  char **v;
  if (i){
    if (!(v=(char **)table_find(i->variables,a))) {
      set(i->parent,a,b);
    } else {
      deallocate_memory(v[1]);
      v[1]=(char *)duplicate_string(b);
      table_insert(i->variables,v,a);
    }
  } else printf ("couldn't set variable %s\n",a);
}

void define(environment i,char *a,char *b)
{
  char **v;
  if (v=(char **)table_find(i->variables,a)){
    deallocate_memory(v[1]);
  } else {
    v=(char **)allocate_memory(2*sizeof(char *));
    v[0]=(char *)duplicate_string(a);
  } 
  v[1]=(char *)duplicate_string(b);
  table_insert(i->variables,v,a);
}

environment create_environment(environment p)
{
  environment new=(environment)allocate_memory(sizeof(struct environment));
  new->variables=create_table(compare_variable,key_from_string);
  new->reference_count=0;
  if (p)
    new->space=create_command_space(p->space);
  else 
    new->space=(command_space)create_command_space(0);
  new->parent=p;
  command_function(new->space,set,"set","","@ss",new);
  command_function(new->space,define,"define","","@ss",new);
  command_function (new->space,skip_if,"if","s","@niss",new);

  command_function(new->space,create_lambda,"lambda","x","@#s",new);
  command_function(new->space,skip_load_file,"load","","@s",new);
  return(new);
}


static call_parser (parser p,char *terminal,
		    char *(*result_function)(),void *result_arg,int flags)
{
  parser_frame new=(parser_frame)allocate_memory(sizeof(struct parser_frame));

  new->p=p;
  new->parent=p->top;
  new->argv[0]=0;
  new->result_function=result_function;
  new->result_arg=result_arg;
  new->flags=flags;
  new->terminal=terminal; 
  new->backslash=0;
  new->fill_pointer=0;
  new->free_length=0;
  new->argc=0;
  p->top=new;
}

static char *increment_argc(parser_frame p,int argc,char **argv)
{
  int i;
  char *z=" ";
  char *argv2[30];
  for (i=0;i<argc;i++)argv2[i]=argv[i];
  argv2[argc]=z;
  append_string(p,argc+1,argv2);
  p->argv[p->argc][p->fill_pointer-1]=0;
  p->argv[++p->argc]=0;
  p->fill_pointer=0;
  p->free_length=0;
  return(0);
}

parser create_skip_parser(environment e,
			  char *(*result_function)(),
			  void *result_arg)
{
  parser new=(parser)allocate_memory(sizeof(struct parser));
  new->result_function=result_function;
  new->result_arg=result_arg;
  new->e=e;
  new->top=0;
  call_parser(new,"",0,0,20);
  return(new);
}

#define EXPAND_VARIABLES        1
#define EVALUATE_SUBEXPRESSIONS 2
#define FORK_ARGUMENT_PARSER    4
#define APPEND_RESULT_TO_PARENT 8
#define FORK_COMMAND_DISPATCH  16
#define THUNK_ARGUMENTS        32
#define DONT_PASS_TERMINAL_UP  64
#define DONT_EVAL_BRACES      128
#define APPEND_TERMINAL       256
#define IN_COMMENT            512

void skip_parse(parser f,char c)
{
  char *result=0,*interim=0,*temp;
  int i;
  char ins;
  char x[2];
  char *z[1];
  parser_frame p=f->top;

  z[0]=x;
  x[0]=c;
  x[1]=0;
  if (f->top->flags&FORK_COMMAND_DISPATCH){
    /* should be the union of new things with the existing terminal*/
    call_parser(f,"\n;",f->result_function,f->result_arg,0);
    call_parser(f,"\n;",command_dispatch,f->e->space,
		FORK_ARGUMENT_PARSER|APPEND_RESULT_TO_PARENT|
		EXPAND_VARIABLES|EVALUATE_SUBEXPRESSIONS);
  } else {
    if (p->backslash) p->backslash=0;
    else if (strchr(p->terminal,c)){
      f->top=p->parent;
      if (p->flags&APPEND_TERMINAL) append_string(p,1,z);
      if (p->argv[p->argc]) increment_argc(p,0,z);
      if (p->result_function)
	z[0]=(*p->result_function)(p->result_arg,p->argc,p->argv);
      /* consider seperating this out into another parsing layer*/
      if (z[0] && (p->flags & APPEND_RESULT_TO_PARENT))
	append_string(p->parent,1,z);
      if (!(p->flags & DONT_PASS_TERMINAL_UP)) skip_parse(f,c);
      for (i=0;i<p->argc+1;i++)
	if (p->argv[i]) deallocate_memory(p->argv[i]);
      deallocate_memory(p);
      return;
    }
  }
  ins=0;
  p=f->top; 
  if (c=='\\'){
    p->backslash^=1;
    if (p->backslash) return;
  }
  if (p->flags&IN_COMMENT) return;
  switch(c){
  case '#': call_parser(f,"\n",0,0,IN_COMMENT);
    break;
  case '{':
    if (p->flags & DONT_EVAL_BRACES){
      call_parser(f,"}",append_string,p,
		  APPEND_TERMINAL|DONT_PASS_TERMINAL_UP|
		  DONT_EVAL_BRACES);
      ins=c;
    } else {
      call_parser(f,"}",append_string,p,
		  DONT_PASS_TERMINAL_UP|DONT_EVAL_BRACES);
      p->flags |=THUNK_ARGUMENTS; /* what about the empty list?*/
    }
    break;
  case '"': call_parser(f,"\"",append_string,p,EXPAND_VARIABLES);
    break;
  case '[': 
    if (p->flags&FORK_ARGUMENT_PARSER){
      call_parser(f," \n\t",append_string,p,
		  EXPAND_VARIABLES|EVALUATE_SUBEXPRESSIONS|
		  FORK_ARGUMENT_PARSER);
      p->flags |=THUNK_ARGUMENTS;
    }
    if (p->flags&EVALUATE_SUBEXPRESSIONS)
      call_parser(f,"]",command_dispatch,f->e->space,
		  APPEND_RESULT_TO_PARENT|EXPAND_VARIABLES|
		  FORK_ARGUMENT_PARSER|DONT_PASS_TERMINAL_UP);
    else ins=c;
    break;
  case '$':
    if (p->flags&EXPAND_VARIABLES){
      call_parser(f," /;\n]}$\"",resolve_variable,f->e,
		  EXPAND_VARIABLES|EVALUATE_SUBEXPRESSIONS|
		  APPEND_RESULT_TO_PARENT);
      p->flags |=THUNK_ARGUMENTS;
    } else ins=c;
    break;
  case ' ':
  case '\t':
  case '\n':
    if (!(p->flags&FORK_ARGUMENT_PARSER)){
      ins=c;
    } else {
      if (p->flags&THUNK_ARGUMENTS) {
	increment_argc(p,0,z);
	p->flags ^=THUNK_ARGUMENTS; /* this is a 'something' flag*/
      }
    }
    break;
  default:
    ins=c;
  }
  if (ins){
    if (p->flags&FORK_ARGUMENT_PARSER){  
      call_parser(f," \n\t]",append_string,p,
		  EVALUATE_SUBEXPRESSIONS|EXPAND_VARIABLES);
      p->flags |= THUNK_ARGUMENTS;
      skip_parse(f,c);
    } else {
      append_string(p,1,z);
    }
  }
}

static list skip_parse_list(char *s,void *context)
{
  list result=0;
  int count=0;
  char *temp;
  int something=0;
  command_type_handler h=(command_type_handler)context;
  do{
    if (!s[count] || (s[count]==' ') || (s[count]=='\n') || (s[count]=='\t')){
      if (something){
	temp=(char *)duplicate_memory(s,count);
	result=insert((*h->conversion_function)(temp,h->context),result);
	/*	deallocate_memory(temp);*/
	something=0;
      }
    } else {
      if (!something) {
	s+=count;
	count=0;
	something=1;
      }
    }
  } while (s[count++]);
  return(result);
}

static char* skip_print_list(list a)
{
  return(0);
}

void deallocate_skip_list(list n,void *context)
{
  command_type_handler h=(command_type_handler)context;
  destroy_list(n,h->deallocate_memory);
}

void skip_list_type (char abbreviation,char *full_name,char *list_of_type)
{
  char zed[2];
  zed[0]=abbreviation;
  zed[1]=0;
  command_type(skip_global_environment->space,zed,full_name,skip_parse_list,
	       skip_print_list,deallocate_skip_list,
	       resolve_type(skip_global_environment->space,list_of_type));
}

void skip_type(char abbreviation,char *full_name,
	       void *(*translation_function)(),
	       char *(*print_function)(),
	       void (*deallocate_memory)())
{
  char zed[2];
  zed[0]=abbreviation;
  zed[1]=0;
  command_type(skip_global_environment->space,zed,full_name,
	       translation_function, print_function, deallocate_memory);
}


void skip_function (void *(*f)(),char *name, char *return_type,
		    char *arguments, void *arg)
{
  command_function(skip_global_environment->space,f,name,return_type,
		   arguments,arg);
}


void skip_variable(void *v, char *name, char *t)
{
  define(skip_global_environment,name,
      command_print_value(skip_global_environment->space,v,t));
}

/** this should return the environment and not use global
  so that one can have multiple top levels */
void initialize_skip()
{
  skip_global_environment=create_environment(0);
  gensym_counter=create_rmap(0,-1);
  command_type (skip_global_environment->space,"i","integer",
		command_parse_decimal,
		command_print_decimal,0);
  command_type (skip_global_environment->space,"f","float",
		command_parse_float,
		command_print_float,deallocate_memory);
  command_type (skip_global_environment->space,"x","hex",
		command_parse_hex,
		command_print_hex,0); /*deallocate_memory);*/
  command_type (skip_global_environment->space,"s","string",
		command_parse_string,
		command_print_string,0);
  /* since parse string just duplicates the pointer */
  command_type (skip_global_environment->space,"@","context",
		command_parse_context,0,0,0);
  command_type (skip_global_environment->space,"n","nargs",
		command_parse_number_of_arguments,0,0,0);
  command_function (skip_global_environment->space,
		    lambda_dispatch,"lambda_dispatch",
		    "s","nxs*",0);
  command_function (skip_global_environment->space,
		    skip_register_timer,"register_timer","i","ix",0);
  skip_list_type ('#',"string_list","s");
  skip_list_type ('$',"float_list","f");
}
