#include <anemone.h>
#include <math.h>
unsigned int default_width=2;
extern int animate;


int xlate_colorspec(char *spec)
{
  unsigned int r,g,b;
  sscanf(spec,"%d %d %d",&r,&g,&b);
  return((r<<16)|(g<<8)|b);
}


void snap(space sp)
{
  (*sp->snap)(sp);
}

static int property_compare(property p,char *name)
{
  return(string_equal(p->name,name));
}

static int property_handle_compare(property_handler h,char *name)
{
  return(string_equal(h->name,name));
}

char *duplicate_string();

static int space_node_lookup(node n,char *name)
{
  return(string_equal(n->name,name));
}

static int space_edge_lookup(edge e,char *name)
{
  return(string_equal(e->name,name));
}


background_line create_background_line(space s,char *color,float *width,
		     list vertices)
{
  unsigned int k=length(vertices);
  float *n=(float *)allocate_memory(k*sizeof(float));
  list i;
  unsigned int j;
  float min_x,max_x,min_y,max_y;
  background_line b=(background_line)
    allocate_memory(sizeof(struct background_line));

  b->width=*width;
  b->c=xlate_colorspec(color);
  b->coordinates=n;
  b->degree=k;
  for (i=vertices,j=0;i;i=i->rest,j++){
    n[j]=*((float *)i->first);
    if (j==1){
      max_x=min_x=n[0];
      min_y=max_y=n[1];
    } else if (j&1){
      min_x=MIN(min_x,n[j-1]);
      max_x=MAX(max_x,n[j-1]);
      min_y=MIN(min_y,n[j]);
      max_y=MAX(max_y,n[j]);
    }
  }
  insert_instrument(s->q,0,s->draw_background_line,0,b,
		    min_x,min_y,
		    max_x,max_y,2);
  return(b);
}


space create_space(void *hardware_local,
		   void (*draw_node)(),
		   void (*draw_edge)(),
		   void (*draw_background_line)(),
		   void (*draw_chart)(),
		   void (*snap)(),
		   int (*redraw)())
{
  space s=(space)allocate_memory(sizeof(struct space));
  s->hardware_local=hardware_local;
  s->draw_node=draw_node;
  s->draw_edge=draw_edge;
  s->draw_background_line=draw_background_line;
  s->snap=snap;
  s->redraw=redraw;
  s->q=create_cell(0);
  s->nodes=create_table(space_node_lookup,key_from_string);
  s->edges=create_table(space_edge_lookup,key_from_string);
  s->default_node_properties=create_table(property_compare,key_from_string);
  s->default_edge_properties=create_table(property_compare,key_from_string);
  s->node_property_handlers=create_table(property_handle_compare,
					 key_from_string);
  s->edge_property_handlers=create_table(property_handle_compare,
					 key_from_string);
  s->xscale=s->yscale=1.0;

  set_node_default(s,"height","50");
  set_node_default(s,"width","75");
  set_node_default(s,"ring_width","3");
  set_node_default(s,"shape","oval");
  set_node_default(s,"color","255 255 255");

  set_edge_default(s,"width","5.0");
  set_edge_default(s,"color","128 128 192");
  return(s);
}

void resize_space()
{
}




