#include <protocol/protocol.h>
#include <elib/timer.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unix.h>

typedef struct filedesc {
  int in,out;
  int blocksize;
  layer l;
  int direction;
  void (*close_function)();
  void *close_context;
} *filedesc;

struct iovec *make_iov(pbuffer b, int *rank)
{
  pbuffer n;
  int i;
  struct iovec *result;

  for (n=b,i=0;n;n=n->next,i++);
  result=(struct iovec *)allocate_memory(sizeof(struct iovec)*i);
  *rank=i;
  for (n=b,i=0;n;n=n->next,i++){
    result[i].iov_base=n->contents;
    result[i].iov_len=n->length;
  }
  return(result);
}

static int fd_read(filedesc new)
{
  pbuffer b=0,d=0;
  int rank;
  struct iovec *v;
  int bytes_available=0;
  int size;
  int result=0;

  if (ioctl(new->in,FIONREAD,&bytes_available)>=0){
    size=MIN(bytes_available,new->blocksize);
    if (size){
      if (new->l->down){
	b=allocate_downwards_buffer(new->l,size);
	/* if rank would be greater than kernel maxium loop w multiple
	   reads*/
	v=make_iov(b,&rank);
	if (readv(new->in,v,rank)==size) {
	  if (size) write_down(new->l,b,0);
	  result=1;
	} else {
	  if (new->l->down) write_down(new->l,0,PROTOCOL_STACK_SHUTDOWN);
	}
	deallocate_memory(v);
      }
      if (new->l->up) {
	b=allocate_buffer(size);
	if ((rank=read(new->in,b->contents,size)) > 0){
	  write_up(new->l,d=allocate_fragment(b,0,rank),0);
	  dereference_buffer(d);
	  result=1;
	} else {
	  result=0;
	  write_up(new->l,0,PROTOCOL_STACK_SHUTDOWN);
	}
      }
    } else {
      result=0;
    }
  }
  if (!result){
    if (new->close_function) (*new->close_function)(new->close_context);
/*    printf ("closing filedesc %d %d\n",new->in,new->out);*/
    if (new->in) close(new->in); /*hmm*/
    if (new->l->up) write_up(new->l,0,PROTOCOL_STACK_SHUTDOWN);
    if (new->l->down) write_down(new->l,0,PROTOCOL_STACK_SHUTDOWN);
    deallocate_memory(new);
  }
  if (b) dereference_buffer(b);      
  return(result);
}

static void fd_write(layer l, pbuffer b, int f)
{
  filedesc fd=(filedesc)l->state;
  int rank;
  if (b){
    struct iovec *v=make_iov(b,&rank);
    writev(fd->out,v,rank);
    deallocate_memory(v);
  }
  if (f & PROTOCOL_CONNECTED){
    if (fd->in >= 0) register_io_handler(fd_read,fd->in,READ_HANDLER,fd);
    if (l->up) write_up(l,0,PROTOCOL_CONNECTED);
  }
}

layer create_filedesc(int in,int out,int blocksize,void (*close_function)(),
		      void *close_context)
     
{
  filedesc new=(filedesc)allocate_memory(sizeof(struct filedesc));
  new->in=in;
  new->out=out;
  new->blocksize=blocksize;
  new->close_function=close_function;
  new->close_context=close_context;
  return(new->l=create_layer(new,fd_write,fd_write,0));
}

