#include <elib/timer.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/nit_buf.h>
#include <protocol/protocol.h>
#include <sys/stropts.h>
#include <fcntl.h>
#ifdef sun
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif

typedef struct nit
{
  int snaplen;
  int tflag;
  int dflag;
  int state;
  int fd;
  layer l;
  int frame_size;
  int total_size;
  int chain_size;
  int index;
  pbuffer chain;
} *nit;

static void nit_from_above(layer l,pbuffer b,int f)
{
  nit n=(nit)l->state;
}

static int nit_read_handler(nit n)
{
  int len,index;
  pbuffer base,temp;
  struct nit_bufhdr hdrp;
  struct nit_iflen nlp;
  struct timeval *tvp = 0;
  struct nit_iftime  ntp;
  struct nit_ifdrops ndp;

  if (ioctl(n->fd,FIONREAD,&len)==0){
    base=allocate_buffer(len);
    read(n->fd, base->contents, len);
    n->chain=append_buffer(n->chain,base);
    dereference_buffer(base);
    n->chain_size+=len;
    while (n->chain_size > n->total_size/*+n->index*/){
      if (!n->frame_size && (n->chain_size > 8)){
	buffer_read(n->chain,n->index,&hdrp,sizeof(hdrp));
	n->index += sizeof (hdrp);
	n->frame_size = hdrp.nhb_msglen;
	n->total_size = hdrp.nhb_totlen;
      }
      if (n->chain_size >= n->total_size){
	temp=allocate_fragment(n->chain,n->index,n->frame_size);
	if (n->state) write_up(n->l,temp,0); 
	dereference_buffer(temp);
	if (n->chain_size-=n->total_size){
	  temp=allocate_fragment(n->chain,n->total_size,n->chain_size);
	  dereference_buffer(n->chain);
	  n->chain=temp;
	} else {
	  n->chain=0;
	  n->chain_size=0;
	}
	n->frame_size=n->total_size=n->index=0;
      }
    }
  }
  return(1);
}


layer create_nit (char *device,int if_flags,int snaplen,int chunksize)
{
  nit n=(nit)allocate_memory(sizeof(struct nit));
  struct strioctl si;
  struct ifreq ifr;
  struct timeval timeout;
  n->state=1;
  if ((n->fd = open("/dev/nit", O_RDONLY)) < 0)
    perror("couldn't open nit");
  n->snaplen=snaplen;
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  ioctl(n->fd, I_SRDOPT, (char *)RMSGD);
  si.ic_timout = INFTIM;
  ioctl(n->fd, I_PUSH, "nbuf");
  si.ic_cmd = NIOCSTIME;
  si.ic_len = sizeof timeout;
  si.ic_dp = (char *)&timeout;
  ioctl(n->fd, I_STR, (char *)&si);
  si.ic_cmd = NIOCSCHUNK;
  si.ic_len = sizeof chunksize;
  si.ic_dp = (char *)&chunksize;
  ioctl(n->fd, I_STR, (char *)&si);
  strncpy(ifr.ifr_name, device, sizeof ifr.ifr_name);
  ifr.ifr_name[sizeof ifr.ifr_name - 1] = '\0';
  si.ic_cmd = NIOCBIND;
  si.ic_len = sizeof ifr;
  si.ic_dp = (char *)&ifr;
  ioctl(n->fd, I_STR, (char *)&si);
  if (if_flags != 0) {
    si.ic_cmd = NIOCSFLAGS;
    si.ic_len = sizeof if_flags;
    si.ic_dp = (char *)&if_flags;
    ioctl(n->fd, I_STR, (char *)&si);
  }
  n->chain=0;
  n->chain_size=0;
  n->frame_size=0;
  n->index=0;
  ioctl(n->fd, I_FLUSH, (char *)FLUSHR);
  register_io_handler(nit_read_handler,n->fd,READ_HANDLER,n);
  return(n->l=create_layer(n,0,0,0));
}
