#include <elib/timer.h>
#include <elib/memory.h>
#include <unix/unix.h>
#include <sys/time.h>
#include <sys/types.h>

static struct timeval beginning_of_epoch;
static int table_width;
static io_handler io_handlers;
static io_handler current;

static int which_io_handler(io_handler j,fd_set fds[3])
{
  if ((FD_ISSET(j->fd,fds) && (j->type & READ_HANDLER)) ||
      (FD_ISSET(j->fd,fds+1) && (j->type & WRITE_HANDLER)) ||
      (FD_ISSET(j->fd,fds+2) && (j->type & EXCEPTION_HANDLER)))
    return ((*j->function)(j->argument));
  return(1);
}

io_handler register_io_handler(int (*function)(), int fd, int type, void *arg)
{
  io_handler i=(io_handler)allocate_memory(sizeof(struct io_handler));
  i->fd=fd;
  i->type=type;
  i->function=function;
  i->argument=arg;
  i->last=&io_handlers;
  if (i->next=io_handlers) i->next->last=&i->next;
  io_handlers=i;
  return(i);
}

void remove_io_handler (io_handler i)
{
  if (*i->last=i->next) i->next->last=i->last;
  deallocate_memory(i);
}

static int timer_block(int interval, int block)
{
  fd_set fds[3];
  struct timeval timeout;
  struct timeval *timeout_pointer;
  unsigned int i,result;
  io_handler j,next,c;

  /* this needs to loop until the entire interval is expired*/
  if (block){
    timeout_pointer=0;
  } else {
    i=interval/1000;
    timeout.tv_sec=(i>0)?i:0;
    i=(interval-(i*1000))*1000;
    timeout.tv_usec=(i>0)?i:0;
    timeout_pointer=&timeout;
  }
  FD_ZERO(fds);
  FD_ZERO(fds+1);
  FD_ZERO(fds+2);
  for (c=io_handlers;c;c=c->next){
    if (c->type & READ_HANDLER) FD_SET(c->fd,fds);
    if (c->type & WRITE_HANDLER) FD_SET(c->fd,fds+1);
    if (c->type & EXCEPTION_HANDLER) FD_SET(c->fd,fds+2);
  }
  if((result=select(table_width,fds,fds+1,fds+2,timeout_pointer))<0){
    printf("timer: select failed");
  }
  if (result > 0)
    for (c=io_handlers;c;c=next){
      next=c->next;
      if ((FD_ISSET(c->fd,fds) && (c->type & READ_HANDLER)) ||
	  (FD_ISSET(c->fd,fds+1) && (c->type & WRITE_HANDLER)) ||
	  (FD_ISSET(c->fd,fds+2) && (c->type & EXCEPTION_HANDLER)))
	if (!(*c->function)(c->argument)){
	  if (*c->last=c->next) c->next->last=c->last;;
	  free(c);
	}
    }
}

static struct timeval beginning_of_epoch;

unsigned int now()
{
  struct timeval result;
  unsigned int delta;
  
  gettimeofday(&result,0);
  delta=1000*(result.tv_sec-beginning_of_epoch.tv_sec)+
    ((result.tv_usec-beginning_of_epoch.tv_usec)/1000);
  return (delta);
}

void init_unix_timer()
{
  io_handlers=0;
  table_width=getdtablesize();
  initialize_timer(timer_block,now);
}
