#include <varargs.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <elib/timer.h>
#include <protocol/protocol.h>
#include <fcntl.h>
#include <unix.h> 

#ifdef  __osf__ 
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif

#include <netinet/in.h>
#include <elib/table.h>

struct sockaddr_in *make_sockaddr_in();

typedef struct udp_state {
  int fd;
  layer l;
  struct sockaddr_in *local;
  table peers;
  void (*new_peer_function)();
  void *new_peer_arg;
} *udp_state;


typedef struct udp_peer {
  layer l;
  udp_state u;
  struct sockaddr_in *peer_address;
  int connected;
} *udp_peer;

static unsigned int key_from_sockaddr_in(struct sockaddr_in *s)
{return (1);}
static unsigned int compare_udp_peer(udp_peer p,struct sockaddr_in *s)
{}

#ifdef IP_ADD_MEMBERSHIP
#define MULTICAST 1
#endif

static void d(layer l,pbuffer b,int f)
{
  int length=buffer_length(b);
  udp_peer p=(udp_peer)l->state;
  
  if (sendto(p->u->fd, b->contents, b->length, 0,
	     (struct sockaddr *)p->peer_address,
	     sizeof(struct sockaddr_in)) != length){
    perror ("sendto failed");
  }
}

static void udp_peer_up (layer l,pbuffer b,int flags)
{
  udp_peer p=(udp_peer)l->state;
  if (flags & PROTOCOL_CONNECTED) p->connected=1;
  write_up(l,b,flags);
}

layer create_udp_peer(layer l,struct sockaddr_in *peer)
{
  udp_state u=(udp_state)l->state;
  udp_peer p=(udp_peer)allocate_memory(sizeof(struct udp_peer));
  p->peer_address=peer;
  p->connected=0;
  p->u=u;
  table_insert(u->peers,p,peer);
  return(p->l=create_layer(p,d,udp_peer_up,0));
}

static int udp_handle_incoming_io(udp_state s)
{
  struct sockaddr_in from;
  int fromsize=sizeof(from);
  int result,pos=0;
  pbuffer b,f;
  int bytes_available;
  udp_peer p;
  layer l;
  
  if (ioctl(s->fd,FIONREAD,&bytes_available)==0){
    b=allocate_buffer(bytes_available);
    while (bytes_available){
      result=recvfrom(s->fd,b->contents+pos,bytes_available,0,
		      (struct sockaddr *)&from,&fromsize);
      if (!(p=table_find(s->peers,&from))){
	(*s->new_peer_function)(&from,l=create_udp_peer(s->l,&from),
				s->new_peer_arg);
	p=(udp_peer)l->state;
	p->l=l;
      }
      if (p->connected) {
	f=allocate_fragment(b,pos,result);
	bytes_available-=result+16;
	pos+=result;
	write_up(p->l,f,0);
	dereference_buffer(f);
      }
    }
    dereference_buffer(b);
  }
  return(1);
}


/*
  if (IN_MULTICAST(ntohl(u->remote->sin_addr.s_addr))) {
    mreq.imr_multiaddr = u->remote->sin_addr;
    mreq.imr_interface.s_addr = INADDR_ANY;
    if (setsockopt(u->fd,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq))<0)
      perror("ip_add_membership");
  }
*/
/*
#ifdef MULTICAST
  if (IN_MULTICAST(ntohl(u->local->sin_addr.s_addr))) {
    if (setsockopt(u->fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl))<0)
      perror("ip_multicast_ttl");
  }
#endif  
*/
/*
#ifdef MULTICAST
  int ttl=1;
  struct ip_mreq mreq;
#endif
*/
/*
/*  if (setsockopt(u->ifd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1)
    perror("reuseaddr");
  if (setsockopt(u->ofd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) == -1)
    perror("reuseaddr");
*/


layer create_udp(struct sockaddr_in *local, void (*new_peer_function)(),
		 void *new_peer_arg)
{
  udp_state u=(udp_state)allocate_memory(sizeof(struct udp_state));
  int one=1;

  u->fd=socket(AF_INET,SOCK_DGRAM,0);
  u->peers=create_table(compare_udp_peer,key_from_sockaddr_in);
  u->new_peer_function=new_peer_function;
  u->new_peer_arg=new_peer_arg;
  u->local=local;
  if (local){
    if (bind (u->fd,(struct sockaddr *)u->local,sizeof(struct sockaddr_in))){
      printf ("binding to port failed\n");
      return(0);
    }
  }
  register_io_handler(udp_handle_incoming_io,u->fd,READ_HANDLER,u);
  return(u->l=create_layer(u,d,0,0));
}
