#include <protocol/protocol.h>
#include <stdarg.h>
#include <elib/timer.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unix.h>

typedef struct uexec {
  int in[2];
  int out[2];
  int blocksize;
  layer l;
  int pid;
  int direction;
} *uexec;

static int fd_read(uexec u)
{
  pbuffer b=0,d=0;
  int rank;
  struct iovec *v;
  int bytes_available=0;
  int size;
  int result=0;

  if (ioctl(u->in[0],FIONREAD,&bytes_available)>=0){
    size=MIN(bytes_available,u->blocksize);
    if (size){
      if (u->l->down){
	b=allocate_downwards_buffer(u->l,size);
	v=make_iov(b,&rank);
	if (readv(u->in[0],v,rank)==size) {
	  if (size) write_down(u->l,b,0);
	  result=1;
	} else {
	  if (u->l->down) write_down(u->l,0,PROTOCOL_STACK_SHUTDOWN);
	}
	deallocate_memory(v);
      }
      if (u->l->up) {
	b=allocate_buffer(size);
	if ((rank=read(u->in[0],b->contents,size)) > 0){
	  write_up(u->l,d=allocate_fragment(b,0,rank),0);
	  dereference_buffer(d);
	  result=1;
	} else {
	  result=0;
	  write_up(u->l,0,PROTOCOL_STACK_SHUTDOWN);
	}
      }
    } else {
      result=0;
    }
  }
  if (b) dereference_buffer(b);      
  return(result);
}

static void u_write(layer l, pbuffer b, int f)
{
  uexec u=(uexec)l->state;
  int rank;
  if (b){
    struct iovec *v=make_iov(b,&rank);
    writev(u->out[1],v,rank);
    deallocate_memory(v);
  }
  if (f & PROTOCOL_CONNECTED){
    register_io_handler(fd_read,u->in[0],READ_HANDLER,u);
    if (l->up) write_up(l,0,PROTOCOL_CONNECTED);
  }
}

layer create_process(char *command,int argcount,...)
{
  char **argv;
  int i;
  uexec new=(uexec)allocate_memory(sizeof(struct uexec));
  va_list a;

  va_start(a,argcount);
  pipe(new->in);
  pipe(new->out);
  new->blocksize=1024;
  argv=(char **)allocate_memory(sizeof(char *)*(argcount+2));
  argv[0]=command;
  argv[argcount+1]=0;
  for (i=1;i<argcount+1;i++)argv[i]=va_arg(a,char *);
  if (!(new->pid=vfork())){
    printf ("executing: ");
    for (i=0;i<argcount+1;i++) printf("%s ",argv[i]);
    printf ("\n");
    dup2(new->in[1],1);
    dup2(new->out[0],0);
    execvp(command,argv);
    perror("exec failed");
  }
  deallocate_memory(argv);
  va_end(a);
  return(new->l=create_layer(new,u_write,u_write,0));
}

