#include <anemone.h>
#include <X11/Xlib.h>
#include <elib/list.h>
#include <elib/memory.h>
#include <elib/machine.h>

typedef struct panel *panel;
typedef struct panel_element *panel_element;
typedef struct xspace *xspace;
typedef struct button *button;
typedef struct text_entry *text_entry;

struct xspace {
  space sp;
  Display *d;
  Screen *s;
  Window w;
  Pixmap p;
  Cursor cr;
  Pixmap crp;
  XFontStruct *f;
  int depth;
  unsigned int plane_mask;
  unsigned black,white,grey,blue,dark_grey,light_grey,dark_dark_grey;
  Colormap c;
  GC node_gc,copy_gc,label_gc,edge_gc,clear_gc,xor_gc,thick_xor_gc;
  GC panel_gc,panel_clear_gc,overscore,underscore,dark_underscore;
  GC arrow_gc,edge_label_gc;
  float x_offset,y_offset;
  float xscale,yscale;
  list current_selection;
  int width,height;
  int current_width;
  int current_node_width;
  int current_edge_color;
  int current_node_color;
  panel right;
  text_entry bottom;
  int background_move;
  int click_x,click_y;
  int background_move_ox,background_move_oy;
  float background_scale;
  cell q;
  unsigned int button_state;
  int rubber_box_x,rubber_box_y;
};

struct panel {
  panel pa;
  xspace s;
  list buttons;
  Pixmap p;
  int x,y;
  int width,height;
  int button_height; 
  int *sizes;
};

struct button {
  panel p;
  Pixmap c;
  Pixmap c_picture;
  char *name;
  int pressed;
  int type;
  int was_pressed;
  int (*frob)();
  void *frob_arg;
};

struct text_entry {
  char *prompt;
  unsigned int content_length;
  char *current_contents;
  int cursor_location;
};

struct panel_element {
  void *context;
  void (*press)();
  void (*release)();
  void (*character)();
  struct panel_element *right;
  struct panel_element *down;
};

panel create_panel();

static _inline_ float xform(xspace s, float x, float y)
{
  return(x*s->xscale*s->sp->xscale+s->x_offset);
}

static _inline_ float yform(xspace s, float x, float y)
{
  return(y*s->yscale*s->sp->yscale+s->y_offset);
}

static _inline_ float xiform(xspace s, int x, int y)
{
  return((x-s->x_offset)/(s->xscale*s->sp->xscale));
}

static _inline_ float yiform(xspace s, int x, int y)
{
  return((y-s->y_offset)/(s->yscale*s->sp->yscale));
}


static _inline_ void draw_text (xspace s,char *string,
				int x,int y,
				int width,int height,
				GC gc)
{
   int len=string_length(string);
   char *used=(char *)allocate_memory(len+4);
   int direction,ascent,descent,i;
   XCharStruct overall;
   bcopy(string,used,len);
   XTextExtents (s->f,string,len,&direction,&ascent,&descent,&overall);
   if (overall.lbearing+overall.rbearing > width){
	  len+=4;
	  used[len-2]=used[len-1]='.';
	  do {
             used[(--len)-2]='.';
	     XTextExtents (s->f,used,len,&direction,&ascent,&descent,&overall);
	  } while ((overall.lbearing+overall.rbearing > width) && (len > 3));
   }
   if ((overall.lbearing+overall.rbearing <= width) &&
	  (overall.ascent+overall.descent <= height))
	  XDrawString(s->d,s->p,gc,
		      x-(overall.lbearing+overall.rbearing)/2,
		      y+ascent/2,
		      used,len);
   deallocate_memory(used);
}

