#include <anemone.h>
#include <x.h>
#include <elib/timer.h>
#include <elib/memory.h>
#include <elib/pqueue.h>
#include <math.h>



static _inline_ void draw_poly_node(space sp,node n)
{
  XPoint p[MAX_POLYGONAL_DEGREE];
  shape sh=n->sh;
  int i;
  xspace s=(xspace)sp->hardware_local;
  int x=xform(s,n->x,n->y);
  int y=yform(s,n->x,n->y);

  for (i=0;i<sh->degree;i++){
    p[i].x=x+n->width*s->xscale*sp->xscale*sh->points[i].x;
    p[i].y=y+n->height*s->yscale*sp->yscale*sh->points[i].y;
  }
  if (s->current_node_color !=n->color){
    XSetForeground(s->d,s->node_gc,x_display_color_from_color(s,n->color));
    s->current_node_color=n->color;
  }
  XFillPolygon(s->d,s->p,n->selected?s->overscore:s->node_gc,
	       p,sh->degree,Convex,CoordModeOrigin);
  /* fix this */
  draw_text(s,n->name,x,y,n->width*s->xscale,n->height*s->yscale,
	    s->label_gc);
}

/* could use draw text here*/
static _inline_ draw_oval_node(space sp,node n)
{
  xspace s=(xspace)sp->hardware_local;
  int direction,ascent,descent,i;
  XCharStruct overall;
  int len=strlen(n->name);
  int tw=(int)MAX((float)n->width*s->xscale,2.0);
  int th=(int)MAX((float)n->height*s->yscale,2.0);
  int x=xform(s,n->x,n->y);
  int y=yform(s,n->x,n->y);
  float btemp;
  int box_x;

  XTextExtents (s->f,n->name,len,&direction,&ascent,&descent,&overall);
  btemp=(float)(overall.ascent+overall.descent)/(float)(s->yscale*n->height);
  btemp*=btemp;
  box_x=(int)((float)n->width*s->xscale*((btemp<1.0)?sqrt(1.0-btemp):0));

  if (s->current_node_color !=n->color){
    XSetForeground(s->d,s->node_gc,x_display_color_from_color(s,n->color));
    s->current_node_color=n->color;
  }
  XFillArc(s->d,s->p,n->selected?s->overscore:s->node_gc,
	   x-tw/2,y-th/2,tw,th,0,360*64);
  draw_text(s,n->name,x,y,box_x,n->height*s->yscale,s->label_gc);
}

static _inline_ draw_ring_node(space sp,node n)
{
  xspace s=(xspace)sp->hardware_local;
  int size=n->ring_width*s->xscale;
  int tw=(int)MAX((float)n->width*s->xscale,2.0);
  int th=(int)MAX((float)n->height*s->yscale,2.0);
  int x=xform(s,n->x,n->y);
  int y=yform(s,n->x,n->y);

  if (s->current_node_color !=n->color){
    XSetForeground(s->d,s->node_gc,x_display_color_from_color(s,n->color));
    s->current_node_color=n->color;
  }
  if (s->current_node_width !=size){
    XSetLineAttributes(s->d,s->node_gc,size,LineSolid,CapButt,JoinMiter);
    s->current_node_width =size;
  }
  XDrawArc(s->d,s->p,n->selected?s->overscore:s->node_gc,
	   x-tw/2,y-th/2,tw,th,0,360*64);
}

void draw_x_node(space sp,node n)
{
  if (n->sh){
    switch(n->sh->kind){
      case oval_shape:
	draw_oval_node(sp,n);
	break;
      case hex_shape:
      case triangle_shape:
      case line_shape:
      case square_shape:
	draw_poly_node(sp,n);
	break;
      case ring_shape:
	draw_ring_node(sp,n);
	break;
    }
  }
}
