#include <anemone.h>
#include <x.h>
#include <unix/unix.h>
#include <elib/timer.h>
#include <elib/memory.h>
#include <elib/pqueue.h>
#include <math.h>
extern draw_x_chart();

typedef struct selection {
  instrument i;
  int ox,oy;
} *selection;

static int allocate_color(xspace s,int red,int green,int blue)
{
  XColor c;
  unsigned long colors[1];
  c.red=red*256;
  c.green=green*256;
  c.blue=blue*256;
  c.flags=DoRed|DoGreen|DoBlue;
  /*  XAllocColorCells(s->d,s->c,0,0,0,colors,1);*/
  if (!XAllocColor(s->d,s->c,&c)) printf ("color error (%d %d %d)\n",
					  red,green,blue);
  /*  c.pixel=s->grey=colors[0];
      XStoreColor(s->d,s->c,&c);
      return(colors[0]);*/
  return(c.pixel);
}

int x_display_color_from_color(xspace s,unsigned int k)
{
  return(allocate_color(s,(k>>16)&255,(k>>8)&255,k&255));
}

static void add_to_current_selection (space sp,instrument i,int order)
{
  xspace s=(xspace)sp->hardware_local;
  selection n;
  if(!((node)i->local)->selected){
    n=(selection)allocate_memory(sizeof(struct selection));
    n->i=i;
    ((node)n->i->local)->selected=order;
    n->ox=((node)n->i->local)->x;
    n->oy=((node)n->i->local)->y;
    s->current_selection=insert(n,s->current_selection);
  }
}

static clear_each (selection n,int order)
{
  if ((((node)n->i->local)->selected) <= order){
    (((node)n->i->local)->selected)=0;
    return (0);
  }
  return(1);
}

static void set_background_color(space sp)
{
  XGCValues my_values;
  int flags= GCForeground|GCBackground|GCFunction|GCPlaneMask;
  xspace s=(xspace)sp->hardware_local;
  if (s->clear_gc) XFreeGC(s->d,s->clear_gc);
  my_values.plane_mask = s->plane_mask;
  my_values.foreground = s->black;
  my_values.background = s->black;
  my_values.function = GXset;
  s->clear_gc=XCreateGC(s->d,s->w,flags,&my_values); 
}

static void refresh_display(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  XCopyArea (s->d,s->p,s->w,s->copy_gc,0,0, s->width,s->height,0,0);
  XCopyArea (s->d,s->right->p,s->w,s->copy_gc, 0,0,
	     s->right->width,s->right->height,s->width,0);
  XFlush (s->d);
}

typedef struct refresh_operation { 
  space sp;
  pqueue q;
} *refresh_operation;

static int redraw_compare(instrument i1,instrument i2)
{
  return (i1->depth > i2->depth);
}

static void local_draw(instrument i,space sp)
{
  (*i->draw)(sp,i->local);
}

static int redraw_display (space sp,int x0,int y0,int x1,int y1,
			   unsigned int depth)
{
  xspace s=(xspace)sp->hardware_local;
  refresh_operation r;
  float xk0,yk0,xk1,yk1;
  float xr0=xform(s,x0,y0)-1;
  float yr0=yform(s,x0,y0)-1;
  float xr1=xform(s,x1,y1)+1;
  float yr1=yform(s,x1,y1)+1;
  
  xr0=MAX(xr0,0);
  yr0=MAX(yr0,0);
  xr1=MIN(xr1,s->width);
  yr1=MIN(yr1,s->height);
  
  /* r=(refresh_operation)allocate_memory(sizeof(struct refresh_operation));
     r->sp=sp;
     r->q=create_pqueue(redraw_compare);*/
  XFillRectangle(s->d,s->p,s->clear_gc,xr0,yr0,xr1-xr0,yr1-yr0);

  xk0=xiform(s,xr0,yr0);
  yk0=yiform(s,xr0,yr0);
  xk1=xiform(s,xr1,yr1);
  yk1=yiform(s,xr1,yr1);
  foreach_instrument_in(sp->q,
			MIN(xk0,xk1),MIN(yk0,yk1),
			MAX(xk0,xk1),MAX(yk0,yk1),
			depth,local_draw,sp);
  /*    destroy_pqueue(r->q);*/
  return(1);
}

static void resize_space(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  if (s->p) XFreePixmap(s->d,s->p);
  s->p=XCreatePixmap (s->d,s->w,s->width,s->height,s->depth);
  redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		 xiform(s,s->width,s->height),
		 yiform(s,s->width,s->height),3);
  refresh_display(sp);
}

static int is_it(selection s,instrument i)
{
  return(s->i==i?0:1);
}

static void update_rubber_box(xspace s,unsigned int x,unsigned int y)
{
  XPoint p[5];
  float x0,x1,y0,y1;
  if (s->current_edge_color !=0xffffff){
    XSetForeground(s->d,s->edge_gc,x_display_color_from_color(s,0xffffff));
    s->current_edge_color=0xffffff;
  }
  if (s->current_width !=3){
    XSetLineAttributes(s->d,s->edge_gc,3,LineSolid,CapButt,JoinMiter);
    s->current_width =3;
  }
  x0=MIN(s->click_x,s->rubber_box_x);
  y0=MIN(s->click_y,s->rubber_box_y);
  x0=MAX(0,x0)-3;
  y0=MAX(0,y0)-3;
  x1=MAX(s->click_x,s->rubber_box_x);
  y1=MAX(s->click_y,s->rubber_box_y);
  x1=MAX(0,x1)+3;
  y1=MAX(0,y1)+3;
  redraw_display(s->sp,xiform(s,x0,y0),yiform(s,x0,y0),
		 xiform(s,x1,y1),yiform(s,x1,y1),3);
  p[0].x=p[4].x=p[3].x=s->click_x;
  p[0].y=p[4].y=p[1].y=s->click_y;
  p[1].x=p[2].x=s->rubber_box_x=x;
  p[2].y=p[3].y=s->rubber_box_y=y;
  XDrawLines(s->d,s->p,s->edge_gc,p,5,CoordModeOrigin);
  refresh_display(s->sp);
}

static void toggle_selection(space sp,instrument i,int type)
{
  xspace s=(xspace)sp->hardware_local;
  if (((node)i->local)->selected) {
    if (type == 2){
      foreach(&s->current_selection,is_it,i);
      ((node)i->local)->selected=0;
    }
  } else {
    add_to_current_selection(sp,i,type);
  }
}

static void rubber_box_toggle(instrument i,space sp)
{
  toggle_selection(sp,i,1);
}

static int display_input_handler(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  XEvent event;
  int count,x,y;
  instrument i;
  list k;
  selection n;
  int redrawp=0;
  float xt,yt;
  float x0,x1,y0,y1;

  while (count=XEventsQueued(s->d,QueuedAfterFlush)){
    while (count--){
      XNextEvent(s->d,&event);    
      switch (event.type){
      case ButtonPress:
	s->button_state=event.xbutton.state;
	x=event.xbutton.x;
	y=event.xbutton.y;
	s->click_x=x;
	s->click_y=y;
	if (event.xbutton.button == 1) {
	  s->rubber_box_x=x;
	  s->rubber_box_y=y;
	}
	/* use quad over this to distinguish things at widget level*/
	if (x>s->width) {
	  for (k=s->right->buttons;(y>s->right->button_height) && k;
	       y-=s->right->button_height,k=k->rest);
	  if (k) press_button(k->first);
	} else {
	  if (i=which_instrument(sp->q,(float)xiform(s,x,y),
				 (float)yiform(s,x,y))){
	    toggle_selection(sp,i,(event.xbutton.button == 2)?1:2);
	    redrawp=1;
	  } else {
	    if (event.xbutton.button == 1){
	      foreach(&s->current_selection,clear_each,3);
	    } else {
	      s->background_move=1;
	      s->background_move_ox=s->x_offset;
	      s->background_move_oy=s->y_offset;
	      s->background_scale=s->xscale;
	    }
	  }
	}
	break;
      case ButtonRelease:
	if (event.xbutton.x>s->width) {
	  y=event.xbutton.y;
	  for (k=s->right->buttons;(y>s->right->button_height) && k;
	       y-=s->right->button_height,k=k->rest);
	  if (k) release_button(k->first);
	}
	if (event.xbutton.button == 1){
	  x0=MIN(s->click_x,s->rubber_box_x);
	  y0=MIN(s->click_y,s->rubber_box_y);
	  x1=MAX(s->click_x,s->rubber_box_x);
	  y1=MAX(s->click_y,s->rubber_box_y);
	  foreach_instrument_in(sp->q,
				xiform(s,x0,y0),
				yiform(s,x0,y0),
				xiform(s,x1,y1),
				yiform(s,x1,y1),
				2,rubber_box_toggle,sp);
	}
	dolist(k,s->current_selection){
	  n=(selection)k->first;
	  n->ox=((node)n->i->local)->x;
	  n->oy=((node)n->i->local)->y;
	}
	dolist(k,s->right->buttons) disarm_button(k->first);
	if (((s->click_x - event.xbutton.x)<10) &&
	    ((s->click_x - event.xbutton.x)>-10) &&
	    ((s->click_y - event.xbutton.y)>-10) &&
	    ((s->click_y - event.xbutton.y)>-10))
	  foreach(&s->current_selection,clear_each,1);
	s->background_move=0;
	redrawp=1;
	break;
      case Expose:
      case VisibilityNotify:
	refresh_display(sp);
	break;
      case MotionNotify:
	if (event.xmotion.state & Button1Mask) 
	  update_rubber_box(s,event.xmotion.x,event.xmotion.y);
	if (event.xmotion.state & Button2Mask) {
	  dolist(k,s->current_selection){
	    n=(selection)k->first;
	    if (n->i->move)
	      (*n->i->move)
		(sp, n->i->local,
		 (float)(n->ox+((event.xmotion.x-s->click_x)/
				s->xscale*sp->xscale)),
		 (float)(n->oy+((event.xmotion.y-s->click_y)/
				s->yscale*sp->yscale)));
	    redrawp=1;
	  }
	  if (s->background_move){
	    s->x_offset=s->background_move_ox+event.xmotion.x-s->click_x;
	    s->y_offset=s->background_move_oy+event.xmotion.y-s->click_y;
	    redrawp=1;
	  }
	}
	if (event.xmotion.state & Button3Mask){
	  xt=(s->x_offset-s->click_x)/s->xscale;
	  yt=(s->y_offset-s->click_y)/s->yscale;
	  s->xscale=s->background_scale+(event.xmotion.y-s->click_y)*0.01;
	  s->yscale=s->xscale=MAX(0.01,s->xscale);
	  s->x_offset=xt*s->xscale+s->click_x;
	  s->y_offset=yt*s->yscale+s->click_y;
	  redrawp=1;
	}
	break;
      case ConfigureNotify:
	s->width=event.xconfigure.width-s->right->width;
	s->right->height=s->height=event.xconfigure.height;
	resize_space(sp);
	x_resize_panel(s->right);
	break;
      }
    }
  }
  if (redrawp){
    redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		   xiform(s,s->width,s->height),
		   yiform(s,s->width,s->height),3);
    refresh_display(sp);
  }
  return(1);
}

x_hack_redraw(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		 xiform(s,s->width,s->height),
		 yiform(s,s->width,s->height),3);
  refresh_display(sp);
}

typedef struct pry {
  int twig;
  float bounds[4];
} *pry;

static void each_node(node n,pry p)
{
  if (p->twig){
    p->bounds[0]=MIN(p->bounds[0],n->x);
    p->bounds[1]=MIN(p->bounds[1],n->y);
    p->bounds[2]=MAX(p->bounds[2],n->x);
    p->bounds[3]=MAX(p->bounds[3],n->y);
  } else {
    p->bounds[0]= p->bounds[2]=n->x;
    p->bounds[1]= p->bounds[3]=n->y;
    p->twig=1;
  }
}

static void autosize (space sp)
{
  xspace s=(xspace)sp->hardware_local;
  struct pry p;
  float bounds[4];
  p.twig=0;
  iterate_table_entries(sp->nodes,each_node,&p);
  if (p.twig){
    s->xscale=s->width/(p.bounds[2]-p.bounds[0]);
    s->yscale=s->height/(p.bounds[3]-p.bounds[1]);
    s->xscale=s->yscale=MIN(s->xscale,s->yscale);
    
    s->x_offset=-p.bounds[0]*s->xscale;
    s->y_offset=-p.bounds[1]*s->yscale;
    
    redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		   xiform(s,s->width,s->height),
		   yiform(s,s->width,s->height),3);
    refresh_display(sp);
  }
}



static void clear_display(space sp)
{
}

static timer autoplace_timer;

static int autoplace_thunk(space sp, int state)
{
  xspace s=(xspace)sp->hardware_local;
  node n;
  list i;
  dolist (i,s->current_selection){
    n=(node)((selection)i->first)->i->local;
    n->fx=0.0;
    n->fy=0.0;
    position_node(n);
  }
  dolist (i,s->current_selection){
    n=(node)((selection)i->first)->i->local;
    move_node(sp,n,(n->x+n->fx),(n->y+n->fy));
  }
  if (s->current_selection){
    redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		   xiform(s,s->width,s->height),
		   yiform(s,s->width,s->height),3);
    refresh_display(sp);
  }
  autoplace_timer=register_timer_wait(1000,autoplace_thunk,sp);
  return(0);
}

static void each_select(node n,space s)
{
  add_to_current_selection(s,n->i,3);
}

static void select_all(space s)
{
  iterate_table_entries(s->nodes,each_select,s);
}

static void toggle_autoplace(space s,int state)
{
  if (state) {
    autoplace_timer=register_timer(100,autoplace_thunk,s);
  } else {
    remove_timer(autoplace_timer);
  }
}

static void call_ungroup(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  list i;
  dolist (i,s->current_selection) 
    ungroup_node(((selection)i->first)->i->local);
}

extern char *x_display_font;

extern int exit();

void draw_x_background_line(space sp,background_line b)
{
  XPoint *p=(XPoint *)allocate_memory(sizeof(XPoint)*(b->degree/2));
  xspace s=(xspace)sp->hardware_local;
  int size=b->width*s->xscale;
  float x,y;
  int i;

  if (s->current_edge_color !=b->c){
    XSetForeground(s->d,s->edge_gc,x_display_color_from_color(s,b->c));
    s->current_edge_color=b->c;
  }
  if (s->current_width !=size){
    XSetLineAttributes(s->d,s->edge_gc,size,LineSolid,CapButt,JoinMiter);
    XSetLineAttributes(s->d,s->arrow_gc,size/4,LineSolid,CapRound,JoinMiter);
    s->current_width =size;
  }
  for (i=0;i<b->degree/2;i++){
    x=b->coordinates[i*2];
    y=b->coordinates[i*2+1];
    p[i].x=xform(s,x,y);
    p[i].y=yform(s,x,y);
  }
   XDrawLines(s->d,s->p,s->edge_gc,p,b->degree/2,CoordModeOrigin);
  deallocate_memory(p);
}

void x_snap(space sp)
{
  xspace s=(xspace)sp->hardware_local;
  redraw_display(sp,xiform(s,0,0),yiform(s,0,0),
		 xiform(s,s->width,s->height),
		 yiform(s,s->width,s->height),3);
  refresh_display(sp);
}

extern void draw_x_node();
extern void draw_x_edge();
extern void dump_startup();

space attach_x_display (char *display_name,int width, int height)
{
  xspace s=(xspace)allocate_memory(sizeof(struct xspace));
  XGCValues my_values;
  XSetWindowAttributes my_attributes;
  int flags= GCForeground|GCBackground|GCFunction|GCPlaneMask;
  
  if (!(s->d=XOpenDisplay(display_name)))
    printf( "X server %s unavailable\n", display_name);
  s->q=create_cell(0);
  s->s=XDefaultScreenOfDisplay(s->d);
  s->c=DefaultColormapOfScreen(s->s);
  s->sp=create_space(s,draw_x_node,draw_x_edge,draw_x_background_line,
		     draw_x_chart,x_snap,
		     redraw_display);
  s->f=XLoadQueryFont(s->d,x_display_font);
  s->depth=DefaultDepthOfScreen(s->s);
  s->plane_mask=(1<<s->depth)-1;
/*BlackPixelOfScreen(s->s); */
  s->black=allocate_color(s,0,0,0);
  s->white=allocate_color(s,255,255,255);
  s->grey=allocate_color(s,105,105,105);
  s->light_grey=allocate_color(s,130,130,130);
  s->dark_grey=allocate_color(s,75,75,75);
  s->dark_dark_grey=allocate_color(s,45,45,45);
  s->blue=allocate_color(s,128,128,192);
  s->height=height;
  s->right=create_panel(s,0,120,s->height,30);
  s->width=width-s->right->width;
  my_attributes.background_pixel=s->black;
  my_attributes.border_pixel=s->black;
  s->w = XCreateWindow(s->d, XDefaultRootWindow(s->d),
		       0,0,width, height, 1,CopyFromParent,
		       InputOutput,CopyFromParent,
		       CWBackPixel|CWBorderPixel,
		       &my_attributes);
  XStoreName(s->d,s->w,"anemone");
  my_values.plane_mask=s->plane_mask;
  my_values.function = GXcopy;
  my_values.background = s->black;
  my_values.foreground = s->white;
  s->node_gc=XCreateGC(s->d,s->w,flags,&my_values); 
  my_values.line_width=s->current_width=5;
  my_values.foreground = s->blue;
  s->edge_gc=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  my_values.foreground = s->dark_grey;
  my_values.font=s->f->fid;
  s->label_gc=XCreateGC(s->d,s->w,flags|GCFont,&my_values); 
  my_values.foreground = s->light_grey;
  my_values.font=s->f->fid;
  s->edge_label_gc=XCreateGC(s->d,s->w,flags|GCFont,&my_values); 
  my_values.background = s->white;
  my_values.foreground = s->black;
  s->copy_gc=XCreateGC(s->d,s->w,flags,&my_values); 
  my_values.background = s->grey;
  my_values.foreground = s->white;
  s->panel_gc=XCreateGC(s->d,s->w,flags|GCFont,&my_values); 
  my_values.background = s->white;
  my_values.foreground = s->grey;
  s->panel_clear_gc=XCreateGC(s->d,s->w,flags|GCFont,&my_values); 
  my_values.line_width=2;
  my_values.foreground = s->light_grey;
  s->overscore=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  my_values.foreground = s->dark_grey;
  s->underscore=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  my_values.foreground = s->dark_dark_grey;
  s->dark_underscore=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  my_values.foreground = s->black;
/*  s->clear_gc=0;*/
  s->clear_gc=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  my_values.line_width = 2;
  s->arrow_gc=XCreateGC(s->d,s->w,flags|GCLineWidth,&my_values); 
  s->p=0;

  s->current_selection=0;
  s->background_move=0;
  s->x_offset=s->y_offset;
  s->xscale=s->yscale=1.0;
  XSelectInput (s->d,s->w,StructureNotifyMask|ExposureMask|
		KeyPressMask|KeyReleaseMask|ButtonPressMask|
		ButtonReleaseMask|KeymapStateMask|Button1MotionMask|
                Button2MotionMask|
		Button3MotionMask|VisibilityChangeMask);
/*  set_background_color(s->sp);*/
  x_resize_panel(s->right);
  resize_space(s->sp);
  XMapWindow(s->d,s->w);  
#if (XlibSpecificationRelease==6)
#define CAST(x) ((_XPrivDisplay)(x))
#else
#define CAST(x) (x)
#endif
  register_io_handler(display_input_handler,CAST(s->d)->fd,
		      READ_HANDLER,s->sp);
  add_button(s->right,"move",0,0,0,0);
  add_button(s->right,"scale",0,0,0,0);
  add_button(s->right,"place",0,0,0,0);
  add_button(s->right,"select all",0,select_all,s->sp,0);
  add_button(s->right,"group",0,0,0,0);
  add_button(s->right,"ungroup",0,call_ungroup,s->sp,0);
  add_button(s->right,"coalesce",0,coalesce,s->sp,0);
  add_button(s->right,"postscript",0,0,0,0);
  add_button(s->right,"xfig",0,0,0,0);
  add_button(s->right,"save startup",0,dump_startup,s->sp,0);
  add_button(s->right,"autoplace",0,toggle_autoplace,s->sp,1);
  add_button(s->right,"autosize",0,autosize,s->sp,0);
  add_button(s->right,"quit",0,exit,0,0);
  return(s->sp);
}
