#include <protocol/protocol.h>
#include <elib/machine.h>

typedef struct blast {
  int xmit_count;
  int recv_count;
  int pdu_size;
  layer l;
  int xmit_size;
  int recv_size;
} *blast;

static int ack(blast f)
{
  int xmit_buffer[65536/4];
  int len=f->xmit_size;
  pbuffer message;
  int i, size;
  unsigned char e;
  while (len) {
    size = (len>f->pdu_size)?f->pdu_size:len;
    len -=size;
    message=allocate_downwards_buffer(f->l,size);
    for (i=1;i<size/4;i++){
      xmit_buffer[i-1]=vhtonl(i);
    }
    buffer_write(message,4,&xmit_buffer,size-4);
    buffer_write_long(message,0,f->xmit_count);
    f->xmit_count++;
    f->xmit_size-=size;
    if (f->l->down) write_down(f->l,message,0);
    if (f->l->up) write_up(f->l,message,0);
    dereference_buffer(message);
  }
}

static void top(layer l, pbuffer b, int flags)
{ 
  int recv_buffer[65536/4];
  blast f=(blast)l->state;
  int length;
  int c;
  int i,k;

  if (b){
    length=buffer_length(b);
    c=buffer_read_long(b,0);
    if (c != f->recv_count){
      printf ("SEQUENCE ERROR got: %d expected: %d\n",c,f->recv_count);
      f->recv_count=c;
      f->recv_count+=32767; /* best guess for loss */
    }
    buffer_read(b,4,&recv_buffer,length-4);
    for (i=1;i<length/4;i++){
      if (vhtonl(recv_buffer[i-1]) != i) {
	printf ("payload error \n");
	i=length;
      }
    }
    f->recv_count++;
    f->recv_size+=length;
    /* printf ("count: %d, bytes=%d\n",f->recv_count, f->recv_size); */
    if ( !f->xmit_size ) {
/*      printf("done: pdu_size =%d\n",f->pdu_size); */
      exit(0);
    }
  
  }
  if (flags & PROTOCOL_CONNECTED){
    register_timer (1,ack,f);
    if (f->l->up) write_up(f->l,0, PROTOCOL_CONNECTED);
  }
}

layer create_blast(int size, int pdu_size)
{
  blast new;
  if (pdu_size>=65536) {
    printf("create_blast: pdu_size too large");
    return(0);
  }
  new=(blast)allocate_memory(sizeof(struct blast));
  new->xmit_size=size;
  new->pdu_size=pdu_size;
  new->recv_size=0;
  new->xmit_count=new->recv_count=0;
  return(new->l=create_layer(new,top,top,0));
}

